/** 
  ******************************************************************************
  *  @file   can_mdr32.c
  *  @brief  Milandr 1986Bx CANx driver for linking with Canfestival library.
  *
  ******************************************************************************
  *  Copyright (C) JSC EREMEX, 2015,2020.
  *  $$LICENSE$
  *****************************************************************************/

#include <MDR32F9Qx_rst_clk.h>
#include <MDR32F9Qx_can.h>
#include <canfestival.h>

enum {    
    MSG_OBJ_MAX = 32,
    MSG_OBJ_TX = 8,
    CAN_DLC_MAX = 8,
};

static void
can_recv_dpc(fx_dpc_t* dpc, void* arg)
{
	fx_canopen_t* can = (fx_canopen_t*)arg;
    MDR_CAN_TypeDef* reg = can->hw_can;
	for (size_t msg_no = MSG_OBJ_TX; msg_no < MSG_OBJ_MAX; msg_no++)
	{
		if (reg->RX & (1 << msg_no))
		{
			if (!(reg->CAN_BUF[msg_no].DLC & CAN_DLC_IDE))
			{
                unsigned int sem_val;
				fx_sem_get_value(&can->sem_rx, &sem_val);

				if (sem_val < COUNT_BUF_CAN_MSG)
				{
					Message* msg = &can->buf_can_rx_msg[can->iw_buf_can_msg];
                    uint32_t id = reg->CAN_BUF[msg_no].ID;
					msg->cob_id = ((id & CAN_ID_SID_Msk) >> CAN_ID_SID_Pos);
                    uint32_t dlc = reg->CAN_BUF[msg_no].DLC;
					msg->len = (dlc & CAN_DLC_Msk) >> CAN_DLC_Pos ;
					msg->rtr = (dlc & CAN_DLC_RTR) >> CAN_DLC_RTR_Pos;

					((uint32_t*)msg->data)[0] = reg->CAN_BUF[msg_no].DATAL;
					((uint32_t*)msg->data)[1] = reg->CAN_BUF[msg_no].DATAH;

					can->iw_buf_can_msg++;
					can->iw_buf_can_msg &= MASK_BUF_CAN_MSG;

					fx_sem_post(&can->sem_rx);
				}
			}

			reg->BUF_CON[msg_no] &= ~(CAN_BUF_CON_RX_FULL);
		}
	}

	return;
}

void
can_recv_isr(void* arg)
{ 
  fx_canopen_t* can_open = (fx_canopen_t*)arg;
  fx_dpc_request(&can_open->dpc_rx, can_recv_dpc, can_open);
}

uint8_t
can_hw_change_baud_rate(MDR_CAN_TypeDef* reg, uint32_t baudrate, uint32_t hclk)
{
	uint32_t clk_bit, div, sjw, pseg, seg1, seg2;
	//
	// Find value with input bitrate
    //
	clk_bit = hclk / baudrate;
	if (clk_bit < 8)
    {
        //
        // Can not find a correct value with input bitrate
        //
        return 2;
    }
	div = clk_bit / 25;
	if ((clk_bit % 25) == 0)
	{
		div -= 1;
		seg1 = 8;
		pseg = 8;
		seg2 = 8;
		sjw  = 4;
	}
	else
	{
		if (div != 0)
			clk_bit = hclk / (baudrate * (div + 1));

		if ((hclk % (baudrate * (div + 1))) < ((baudrate * (div + 1)) / 2))
			clk_bit -= 1;

		seg1 = clk_bit / 3;
		seg2 = (clk_bit - seg1) / 2;
		pseg = clk_bit - seg1 - seg2;
		sjw  = seg2 / 2;
	}
    
	reg->BITTMNG = ((0 << CAN_BITTMNG_SB_Pos) |
                    (sjw << CAN_BITTMNG_SJW_Pos) |
                    ((seg2 - 1) << CAN_BITTMNG_SEG2_Pos) |
                    ((seg1 - 1) << CAN_BITTMNG_SEG1_Pos) |
                    ((pseg - 1) << CAN_BITTMNG_PSEG_Pos) |
                    (div << CAN_BITTMNG_BRP_Pos));

	return 0;
}


uint8_t
can_hw_send(MDR_CAN_TypeDef* reg, Message* msg)
{
    	uint8_t res = 1;
    	for (size_t i = 0; i < MSG_OBJ_TX; i++)
	{
		if (reg->TX & (1 << i))
		{
			uint32_t len = msg->len;
			if (len > CAN_DLC_MAX) len = CAN_DLC_MAX;

            uint32_t id = ((msg->cob_id << CAN_ID_SID_Pos) & CAN_ID_SID_Msk);
			reg->CAN_BUF[i].ID = id;

			uint32_t mdr_dlc = (len << CAN_DLC_Pos);
			if (msg->rtr) mdr_dlc |= CAN_DLC_RTR;

			reg->CAN_BUF[i].DLC = mdr_dlc;

			reg->CAN_BUF[i].DATAL = ((uint32_t*)msg->data)[0];
			reg->CAN_BUF[i].DATAH = ((uint32_t*)msg->data)[1];
			reg->BUF_CON[i] |= CAN_BUF_CON_TX_REQ;
			res = 0;
			break;
		}
	}
    return res;
}


uint8_t
can_hw_init(MDR_CAN_TypeDef* reg, uint32_t baudrate, uint32_t hclk)
{
    uint32_t i;
    reg->CONTROL &= ~(CAN_CONTROL_CAN_EN);

    //
	// CLOCK CAN
    //
	if (reg == MDR_CAN1)
	{
		RST_CLK_PCLKcmd(RST_CLK_PCLK_CAN1, ENABLE);
		MDR_RST_CLK->CAN_CLOCK &= ~(RST_CLK_CAN_CLOCK_CAN1_CLK_EN
                                    | RST_CLK_CAN_CLOCK_CAN1_BRG_Msk);
		MDR_RST_CLK->CAN_CLOCK |= (RST_CLK_CAN_CLOCK_CAN1_CLK_EN
                                    | (0x00 << RST_CLK_CAN_CLOCK_CAN1_BRG_Pos));
	}
	else if (reg == MDR_CAN2)
	{
		RST_CLK_PCLKcmd(RST_CLK_PCLK_CAN2, ENABLE);
		MDR_RST_CLK->CAN_CLOCK &= ~(RST_CLK_CAN_CLOCK_CAN2_CLK_EN
                                    | RST_CLK_CAN_CLOCK_CAN2_BRG_Msk);
		MDR_RST_CLK->CAN_CLOCK |= (RST_CLK_CAN_CLOCK_CAN2_CLK_EN
                                    | (0x00 << RST_CLK_CAN_CLOCK_CAN2_BRG_Pos));
	}
	else
	{
		return 1;
	}
    
    can_hw_change_baud_rate(reg, baudrate, hclk);
    //
	// Basic Mode
    //
	reg->CONTROL = 0;

	for (i = 0; i < 32; i++)
	{
		reg->CAN_BUF_FILTER[i].FILTER = 0;
		reg->CAN_BUF_FILTER[i].MASK = 0;

	}

    //
	// Transfer data to message RAM
    //
	for (i = 0; i < MSG_OBJ_TX; i++)
	{
		reg->CAN_BUF[i].ID = 0;
		reg->CAN_BUF[i].DLC = 0;
		reg->CAN_BUF[i].DATAL = 0;
		reg->CAN_BUF[i].DATAH = 0;
		reg->BUF_CON[i] = CAN_BUF_CON_EN;
	}

    //
	// Receiver data to message RAM
    //
	for (; i < MSG_OBJ_MAX; i++)
	{
		reg->CAN_BUF[i].ID = 0;
		reg->CAN_BUF[i].DLC = 0;
		reg->CAN_BUF[i].DATAL = 0;
		reg->CAN_BUF[i].DATAH = 0;

		reg->BUF_CON[i] = CAN_BUF_CON_RX_TXN | CAN_BUF_CON_EN;
	}
    
    

	CAN_ITConfig(reg, CAN_IT_RXINTEN | CAN_IT_GLBINTEN, ENABLE);
	CAN_RxITConfig(reg, 0xFFFFFFFF, ENABLE);
	CAN_Cmd(reg, ENABLE);
    
    return 0;
}
