#include <stdint.h>
#include <cm3/canfestival.h>
#include <timer.h>
#include <can.h>

#include <stm32f1xx_hal.h>
#include "stm32_defs.h"

static CAN_HandleTypeDef     CanHandle;

/* Gestion d'un buffer tournant de messages CAN en Tx pour que le transfert
 * en mode block fonctionne avec SDO_BLOCK_SIZE > 3
 */
#define MESS_BUFFER_SIZE (SDO_BLOCK_SIZE + 3)
typedef struct
{
	Message Buf[MESS_BUFFER_SIZE];
	Message* readp;
	Message* writep;
} MessBuf_t;
static MessBuf_t messBuf;
void MessBuf_Init(MessBuf_t* pFifo);
bool MessBuf_Write(MessBuf_t* pFifo, Message* pgd);
bool MessBuf_Read(MessBuf_t* pFifo, Message* pgd);

static bool tx_fifo_in_use = false;

static CO_Data* co_data = NULL;

/* prescaler values for 87.5%  sampling point (88.9% at 1Mbps)
   if unknown bitrate default to 50k
*/
static uint16_t
brp_from_birate(uint32_t bitrate)
{
	if (bitrate == 10000)
		return 225;
	if (bitrate == 50000)
		return 45;
	if (bitrate == 125000)
		return 18;
	if (bitrate == 250000)
		return 9;
	if (bitrate == 500000)
		return 9;
	if (bitrate == 1000000)
		return 4;
	return 45;
}

/**
  * @brief CAN MSP Initialization
  *        This function configures the hardware resources used in this example:
  *           - Peripheral's clock enable
  *           - Peripheral's GPIO Configuration
  *           - NVIC configuration for DMA interrupt request enable
  * @param hcan: CAN handle pointer
  * @retval None
  */
void
HAL_CAN_MspInit(CAN_HandleTypeDef* hcan)
{
	/* CAN1 Periph clock enable */
	CANx_CLK_ENABLE();
	/* Enable GPIO clock ****************************************/
	CANx_GPIO_CLK_ENABLE();
	/* Enable AFIO clock and Remap CAN PINs to PB8 and PB9*******/
	CANx_AFIO_REMAP_CLK_ENABLE();
	CANx_AFIO_REMAP_RX_TX_PIN();

	/* CAN1 TX GPIO pin configuration */
	GPIO_InitTypeDef GPIO_InitStruct;
	GPIO_InitStruct.Pin = CANx_TX_PIN;
	GPIO_InitStruct.Mode = GPIO_MODE_AF_PP;
	GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_HIGH;
	GPIO_InitStruct.Pull = GPIO_PULLUP;
	HAL_GPIO_Init(CANx_TX_GPIO_PORT, &GPIO_InitStruct);
	/* CAN1 RX GPIO pin configuration */
	GPIO_InitStruct.Pin = CANx_RX_PIN;
	GPIO_InitStruct.Mode = GPIO_MODE_INPUT;
	GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_HIGH;
	GPIO_InitStruct.Pull = GPIO_PULLUP;
	HAL_GPIO_Init(CANx_RX_GPIO_PORT, &GPIO_InitStruct);

	/* NVIC configuration for CAN1 Reception complete interrupt */
	HAL_NVIC_SetPriority(CANx_RX_IRQn, 0, 0);
	HAL_NVIC_EnableIRQ(CANx_RX_IRQn);

	HAL_NVIC_SetPriority(CANx_TX_IRQn, 0, 0);
	HAL_NVIC_EnableIRQ(CANx_TX_IRQn);
}

/**
  * @brief CAN MSP De-Initialization
  *        This function frees the hardware resources used in this example:
  *          - Disable the Peripheral's clock
  *          - Revert GPIO to their default state
  * @param hcan: CAN handle pointer
  * @retval None
  */
void HAL_CAN_MspDeInit(CAN_HandleTypeDef* hcan)
{
	/*##-1- Reset peripherals ##################################################*/
	CANx_FORCE_RESET();
	CANx_RELEASE_RESET();

	/*##-2- Disable peripherals and GPIO Clocks ################################*/
	/* De-initialize the CAN1 TX GPIO pin */
	HAL_GPIO_DeInit(CANx_TX_GPIO_PORT, CANx_TX_PIN);
	/* De-initialize the CAN1 RX GPIO pin */
	HAL_GPIO_DeInit(CANx_RX_GPIO_PORT, CANx_RX_PIN);

	/*##-4- Disable the NVIC for CAN reception #################################*/
	HAL_NVIC_DisableIRQ(CANx_RX_IRQn);
}

/**
  * @brief  This function is executed in case of error occurrence.
  * @param  None
  * @retval None
  */
static void Error_Handler(void)
{
	while (1)
	{
	}
}

bool
canInit(CO_Data* d, uint32_t bitrate)
{
	/* save the canfestival handle */
	co_data = d;
	MessBuf_Init(&messBuf);

	/* Configure the CAN peripheral */
	CanHandle.Instance = CANx;

	CanHandle.Init.TimeTriggeredMode = DISABLE;
	CanHandle.Init.AutoBusOff = DISABLE;
	CanHandle.Init.AutoWakeUp = DISABLE;
	CanHandle.Init.AutoRetransmission = ENABLE;
	CanHandle.Init.ReceiveFifoLocked = DISABLE;
	CanHandle.Init.TransmitFifoPriority = DISABLE;
	CanHandle.Init.Mode = CAN_MODE_NORMAL;
	/* CAN Baudrate (CAN clocked at 36 MHz)  36e6 / ( prescaler * (1+BS1+BS2))  */
	CanHandle.Init.SyncJumpWidth = CAN_SJW_1TQ;
	if (bitrate == 1000000)
	{
		CanHandle.Init.TimeSeg1 = CAN_BS1_7TQ;
		CanHandle.Init.TimeSeg2 = CAN_BS2_1TQ;
	}
	else if (bitrate == 500000)
	{
		CanHandle.Init.TimeSeg1 = CAN_BS1_6TQ;
		CanHandle.Init.TimeSeg2 = CAN_BS2_1TQ;
	}
	else
	{
		CanHandle.Init.TimeSeg1 = CAN_BS1_13TQ;
		CanHandle.Init.TimeSeg2 = CAN_BS2_2TQ;
	}
	CanHandle.Init.Prescaler = brp_from_birate(bitrate);


	if (HAL_CAN_Init(&CanHandle) != HAL_OK)
	{
		/* Initialization Error */
		Error_Handler();
	}

	/* Configure the CAN Filter */
	CAN_FilterTypeDef sFilterConfig = {
		.FilterBank = 0,
		.FilterMode = CAN_FILTERMODE_IDMASK,
		.FilterScale = CAN_FILTERSCALE_32BIT,
		.FilterIdHigh = 0x0000,
		.FilterIdLow = 0x0000,
		.FilterMaskIdHigh = 0x0000,
		.FilterMaskIdLow = 0x0000,
		.FilterFIFOAssignment = CAN_RX_FIFO0,
		.FilterActivation = ENABLE
		//.SlaveStartFilterBank = 14;
	};

	/* Enable FIFO 0 message pending Interrupt */
	if (HAL_CAN_ActivateNotification(&CanHandle, CAN_IT_RX_FIFO0_MSG_PENDING) != HAL_OK)
	{
		Error_Handler();
	}

	if (HAL_CAN_ConfigFilter(&CanHandle, &sFilterConfig) != HAL_OK)
	{
		Error_Handler();
	}

	if (HAL_CAN_Start(&CanHandle) != HAL_OK)
	{
		Error_Handler();
	}

	return true;
}

void canClose(void)
{
	HAL_NVIC_DisableIRQ(CAN1_TX_IRQn);
	HAL_NVIC_DisableIRQ(CAN1_RX0_IRQn);
	HAL_CAN_DeInit(&CanHandle);
	__HAL_RCC_CAN1_CLK_DISABLE();
}

//
// The driver send a CAN message passed from the CANopen stack
bool lowlevel_canSend(Message* m)
{
	CAN_TxHeaderTypeDef TxHeader;
	TxHeader.StdId = m->cob_id;
	TxHeader.IDE = CAN_ID_STD;
	TxHeader.RTR = (m->rtr) ? CAN_RTR_REMOTE : CAN_RTR_DATA;
	TxHeader.DLC = m->len;

	uint8_t TxData[8];
	for (size_t i = 0 ; i < m->len ; i++)
	{
		TxData[i] = m->data[i];
	}

	uint32_t TxMailbox;
	HAL_StatusTypeDef status;
	status = HAL_CAN_AddTxMessage(&CanHandle, &TxHeader, TxData, &TxMailbox);

	return (status == HAL_OK) ? true : false;
}

void
USB_HP_CAN1_TX_IRQHandler(void)
{
	Message msg;
	__HAL_CAN_CLEAR_FLAG(&CanHandle, CAN_FLAG_RQCP0);
	__HAL_CAN_CLEAR_FLAG(&CanHandle, CAN_FLAG_RQCP1);
	__HAL_CAN_CLEAR_FLAG(&CanHandle, CAN_FLAG_RQCP2);

	if (!MessBuf_Read(&messBuf, &msg))
	{
		HAL_CAN_DeactivateNotification(&CanHandle, CAN_IT_TX_MAILBOX_EMPTY); //empty fifo, no frames to send
		tx_fifo_in_use = false;
	}
	else
	{
		lowlevel_canSend(&msg);
	}
}

// The driver send a CAN message passed from the CANopen stack
unsigned char
canSend(CAN_PORT notused, Message* m)
{
	bool rv;
	HAL_CAN_DeactivateNotification(&CanHandle, CAN_IT_TX_MAILBOX_EMPTY);
	if (tx_fifo_in_use)
	{
		rv = MessBuf_Write(&messBuf, m) ? 0 : 1;
	}
	else
	{
		rv = lowlevel_canSend(m);
		//
		// if no free mailbox we write in fifo
		//
		if (!rv)
		{
			rv = MessBuf_Write(&messBuf, m) ? 0 : 1;
			tx_fifo_in_use = true;
		}
	}

	if (tx_fifo_in_use)
	{
		HAL_CAN_ActivateNotification(&CanHandle, CAN_IT_TX_MAILBOX_EMPTY);
	}

	return rv;
}

//The driver pass a received CAN message to the stack
/*
unsigned char canReceive(Message *m)
{
}
*/
unsigned char canChangeBaudRate_driver(CAN_HANDLE fd, char* baud)
{
	return 0;
}

/**
  * @brief  This function handles CAN1 RX0 interrupt request.
  * @param  None
  * @retval None
  */
void
USB_LP_CAN1_RX0_IRQHandler(void)
{
	CAN_RxHeaderTypeDef RxHeader;
	uint8_t RxData[8];
	HAL_CAN_GetRxMessage(&CanHandle, CAN_RX_FIFO0, &RxHeader, RxData);
	// Drop extended frames
	if (RxHeader.IDE == CAN_ID_EXT)
	{
		return;
	}

	Message rxm = {0};
	rxm.cob_id = RxHeader.StdId;
	if (RxHeader.RTR == CAN_RTR_REMOTE)
	{
		rxm.rtr = 1;
	}
	rxm.len = RxHeader.DLC;
	for (size_t i = 0 ; i < rxm.len ; i++)
	{
		rxm.data[i] = RxData[i];
	}

	canDispatch(co_data, &rxm);
}

void
MessBuf_Init(MessBuf_t* pFifo)
{
	pFifo->readp = pFifo->writep = pFifo->Buf;
}

bool
MessBuf_Write(MessBuf_t* pFifo, Message* pgd)
{
	int Space = (pFifo->writep < pFifo->readp) ?
	        (pFifo->readp - pFifo->writep - 1) :
			(pFifo->readp - pFifo->writep + (MESS_BUFFER_SIZE - 1));
	if (Space == 0)
		return 1;
	*(pFifo->writep) = *pgd;
	pFifo->writep++;
	if (pFifo->writep >= (pFifo->Buf + MESS_BUFFER_SIZE))
		pFifo->writep = pFifo->Buf;
	return 0;
}

bool MessBuf_Read(MessBuf_t* pFifo, Message* pgd)
{
	bool have_data = (pFifo->readp == pFifo->writep) ? false : true;
	if (have_data)
	{
		*pgd = *(pFifo->readp);
		pFifo->readp++;
		if (pFifo->readp >= (pFifo->Buf + MESS_BUFFER_SIZE))
		{
			pFifo->readp = pFifo->Buf;
		}
	}
	return have_data;
}
