
//#include <board.h>
//#include "dma/dma.h"
//#include "intmath.h"
//#include "irq/irq.h"
#include <sdmmc_api.h>
#include "hsmci.h"
#include "hsmcid.h"
#include "hsmci_dma.h"

#include <system.h>

/**
 * \brief Prepare DMA channels
 * \return SDMMC_SUCCESS if the DMA channels are ready for transfer setup.
 */
uint8_t hsmci_prepare_dma_channels(struct _hsmci_set *set)
{
	set->dma_rx_ch_num = 0;
	set->dma_tx_ch_num = 1;
	return SDMMC_SUCCESS;
}

/**
 * \brief Configure and start DMA for the specified data transfer
 * \return SDMMC_SUCCESS if the DMA transfer has been started.
 */
#define DMA_DATA_WIDTH_BYTE        0
#define DMA_DATA_WIDTH_HALF_WORD   1
#define DMA_DATA_WIDTH_WORD        2

/** Max DMA single transfer size */
#define DMAC_MAX_BT_SIZE 0xFFFF

/** Max DMA length of the block  */
#define DMAC_MAX_BLOCK_LEN 0xFFFF

uint8_t dma_cfg_single_block_read(struct _hsmci_set *set, Dmac* dmac)
{
	sSdmmcCommand * const cmd = set->cmd;

	dmac->DMAC_EN = 1;
	
	/* 2. Clear any pending interrupts on the channel from the previous DMA transfer by reading the DMAC_EBCISR. */
	uint32_t pending_irqs = dmac->DMAC_EBCISR;
	DPRINTF(DBG_INFO, "pending_irqs=%x", pending_irqs);

    /* 3. Program the channel registers.*/
	DmacCh* rx_ch = &dmac->DMAC_CH[set->dma_rx_ch_num];
	DmacCh* tx_ch = &dmac->DMAC_CH[set->dma_tx_ch_num];

	/* RX Channel */
	/* The DMAC_SADDRx for Channel x must be set with the starting address of the HSMCI_FIFO address */
	rx_ch->DMAC_SADDR = set->regs->HSMCI_FIFO;
	rx_ch->DMAC_DADDR = fx_task_get_memaddr((uintptr_t)cmd->pData);

	/* Configure data count */
	size_t btsize = (cmd->wBlockSize / 4);// * (uint32_t)cmd->wNbBlocks;

	rx_ch->DMAC_CTRLA = (DMA_DATA_WIDTH_WORD << DMAC_CTRLA_SRC_WIDTH_Pos)
		| (DMA_DATA_WIDTH_WORD << DMAC_CTRLA_DST_WIDTH_Pos)
		| (DMAC_CTRLA_SCSIZE_CHK_16)
		| (DMAC_CTRLA_DCSIZE_CHK_16)
		| btsize;

/*
	–FIFOCFG defines the watermark of the DMA channel FIFO.
	–SRC_H2SEL is set to true to enable hardware handshaking on the destination.
	–SRC_PER is programmed with the hardware handshaking ID of the targeted HSMCI Host Controller.
	–Enable Channel x, writing one to DMAC_CHER[x]. The DMAC is ready and waiting for request.
*/
	rx_ch->DMAC_CFG =
		  DMAC_CFG_SRC_PER(set->id) 	// HSMCI Peripheral id
		| DMAC_CFG_SRC_PER_MSB(set->id) // HSMCI Peripheral id
		| DMAC_CFG_SRC_REP_CONTIGUOUS_ADDR
		| DMAC_CFG_DST_REP_CONTIGUOUS_ADDR
		| DMAC_CFG_SRC_H2SEL_HW
		| DMAC_CFG_FIFOCFG_ASAP_CFG;
		
/*
	–DST_INCR is set to INCR.
	–SRC_INCR is set to INCR.
	–FC field is programmed with peripheral to memory flow control mode.
	–Both DST_DSCR and SRC_DSCR are set (descriptor fetch is disabled).
	–DIF and SIF are set with their respective layer ID. If SIF is different from DIF, the DMA controller is
	able to prefetch data and write HSMCI simultaneously.
*/
	rx_ch->DMAC_CTRLB = DMAC_CTRLB_SRC_INCR_INCREMENTING | DMAC_CTRLB_DST_INCR_INCREMENTING;
	rx_ch->DMAC_CTRLB |= DMAC_CTRLB_FC_PER2MEM_DMA_FC;
	rx_ch->DMAC_CTRLB |= DMAC_CTRLB_SRC_DSCR_FETCH_DISABLE | DMAC_CTRLB_DST_DSCR_FETCH_DISABLE;
	rx_ch->DMAC_CTRLB |= DMAC_CTRLB_SIF_AHB_IF2;
	rx_ch->DMAC_CTRLB |= DMAC_CTRLB_DIF_AHB_IF0;

	rx_ch->DMAC_DSCR = 0;
	dmac->DMAC_WPMR = 0;

	fx_task_cache_control(TASK_CACHE_INVAL, (uintptr_t)cmd->pData, cmd->wBlockSize);

	DPRINTF(DBG_INFO, "DMAC Enable Channel!");	
	dmac->DMAC_CHER = DMAC_CHER_ENA0 << set->dma_rx_ch_num;

	return SDMMC_SUCCESS;
}

uint8_t hsmci_configure_dma(struct _hsmci_set *set)
{
	switch (set->cmd->bCmd)
	{
		case 17:
			dma_cfg_single_block_read(set, set->dmac);
			break;
		default:
			break;
	}
	return SDMMC_SUCCESS;
}