/**
****************************************************************************************************
*  @file   sdmmc.c
*  @brief  SD/MMC controller and device driver.
*
****************************************************************************************************
*  Copyright (C) 2019-2020 Eremex
*  All rights reserved.
***************************************************************************************************/

#include <system.h>

#include <hsmci.h>
#include <hsmcid.h>

#include "board_support.h"
#include "sdmmc.h"

static bool
init_device(sSdCard *pSd)
{
	uint8_t rc = SD_Init(pSd);
	if (rc != SDMMC_OK) {
		return false;
	}
	return true;
}

static bool
deinit_device(sSdCard *pSd)
{
	SD_DeInit(pSd);
	return true;
}

void*
sdmmc_open(void* arg, uint32_t tid, int32_t flags)
{
    return arg;
}

int
sdmmc_close(void* arg, uint32_t tid)
{
    return 0;
}

int32_t
sdmmc_read(void* arg, uint32_t tid, void *buf, uint64_t block_start, int32_t nbytes)
{
    sdmmc_t* sdmmc = arg;
	DPRINTF(DBG_INFO, "off=%d, nbytes=%d", block_start, nbytes);
    int32_t nblocks = nbytes / 512; // TODO bytes => blocks properly
    int32_t result = SD_ReadBlocks(&sdmmc->card, block_start/512, buf, nblocks);
	if (result != SDMMC_OK)
	{
			return 0;
	}

    return nbytes;
}

int32_t
sdmmc_write(void* arg, uint32_t tid, const void* buf, uint64_t block_start, int32_t nbytes)
{
	sdmmc_t* sdmmc = arg;
	DPRINTF(DBG_INFO, "off=%d, nbytes=%d", block_start, nbytes); // TODO bytes => blocks properly
	int32_t nblocks = nbytes / 512; // TODO bytes => blocks properly
    int32_t result = SD_WriteBlocks(&sdmmc->card, block_start/512, buf, nblocks);
	if (result != SDMMC_OK)
	{
		return 0;
	}

    return nbytes;
}

void*
sdmmc_init(void* sys_args, const char* cfg)
{
	int rc;
	DPRINTF(DBG_INFO, "Initializing");
	sdmmc_t* sdmmc = malloc(sizeof(sdmmc_t));
	if (!sdmmc)
	{
		DPRINTF(DBG_ERROR, "No space for driver data");
		goto end;
	}
	memset(sdmmc, 0, sizeof(*sdmmc));
	sSdCard *card = &sdmmc->card;
	struct _hsmci_set* set = &sdmmc->settings;

	if (false == board_init(cfg, set))
	{
		DPRINTF(DBG_ERROR, "Board specific initialization failed");
		goto end;
	}

	char cfg_path[64];
    strncpy(cfg_path, cfg, strlen(cfg) + strlen("use_irq") + 1);
    strcat(cfg_path, "use_irq");
    if (fx_reg_get_value(cfg_path, FX_REG_TYPE_DWORD, &set->use_irq, sizeof(set->use_irq)))
	{
		set->use_irq = 0;
	}
    strncpy(cfg_path, cfg, strlen(cfg) + strlen("use_dma") + 1);
    strcat(cfg_path, "use_dma");
    if (fx_reg_get_value(cfg_path, FX_REG_TYPE_DWORD, &set->use_dma, sizeof(set->use_dma)))
	{
		set->use_dma = 0;
	}
	//set->dma_unlocks_mutex = 0;

	sSdHalFunctions* sdHal = hsmci_initialize(&sdmmc->settings);
	SDD_Initialize(&sdmmc->card, &sdmmc->settings, sdHal);

	if (SD_GetStatus(card) == SDMMC_NOT_SUPPORTED)
	{
		DPRINTF(DBG_ERROR, "SD card not detected");
		free(sdmmc);
		sdmmc = NULL;
		goto end;
	}
		
	if (!init_device(card))
	{
		DPRINTF(DBG_ERROR, "Couldn't open SD card");
		free(sdmmc);
		sdmmc = NULL;
		goto end;
	}
	DPRINTF(DBG_INFO, "OK");
end:
	return sdmmc;
}

void
sdmmc_deinit(void* arg)
{
	;
}
