/**
****************************************************************************************************
*  @file   system.h
*  @brief  OS and compiler dependencies for driver.
*
****************************************************************************************************
*  Copyright (C) 2019-2020 Eremex
*  All rights reserved.
***************************************************************************************************/

#ifndef _SYSTEM_H_
#define _SYSTEM_H_

#include <stdint.h>
#include <stdbool.h>
#include <string.h>

#include <native/FXRTOS_LIB.h>
#include <native/fx_libc.h>
#include <native/ldr.h>
#include <native/debug.h>

#define __I  volatile const   /**< Defines 'read-only'  permissions */
#define   __O  volatile	      /**< Defines 'write-only' permissions */
#define   __IO volatile	      /**< Defines 'read/write' permissions */

typedef fx_mutex_t mutex_t;

#define assert(a) true
#define trace_error(...) do { ; } while (0)
#define trace_info(...) do { ; } while (0)
#define trace_debug(...) do { ; } while (0) //DPRINTF(DBG_INFO, __VA_ARGS__)
#define trace_warning(...) do { ; } while (0)

#define trace_fatal(...) \
    do { ; while (1) ; } while (0)


#define TICK_PERIOD     1000 // 1ms

struct _timeout {
    uint32_t    start;
    uint32_t    count;
};

void timer_start_timeout(struct _timeout *timeout, uint32_t usec);

bool timer_timeout_reached(struct _timeout* timeout);

void sleep_ms(uint32_t msec);

void sleep_us(uint32_t usec);

bool irq_add_handler(uint32_t periph_id, void(*isr)(void*), void* isr_arg);

bool irq_enable(uint32_t periph_id);

/**
 *  Returns the minimum value between two integers.
 *  \param a First integer to compare
 *  \param b Second integer to compare
 */
static inline uint32_t min_u32(uint32_t a, uint32_t b)
{
    return a < b ? a : b;
}

/**
 *  Returns the maximum value between two integers.
 *  \param a First integer to compare
 *  \param b Second integer to compare
 */
static inline uint32_t max_u32(uint32_t a, uint32_t b)
{
    return a > b ? a : b;
}

/**
 *  Returns the absolute value of an integer.
 *  \param value Integer value
 */
static inline uint32_t abs_u32(int32_t value)
{
    return value > 0 ? value : -value;
}

/**
 *  Computes and returns x to the power of y.
 *  \param x Value
 *  \param y Power
 */
static inline uint32_t power_u32(uint32_t x, uint32_t y)
{
        uint32_t result = 1;
        while (y > 0) {
                result *= x;
                y--;
        }
        return result;
}

/** ISO/IEC 14882:2003(E) - 5.6 Multiplicative operators:
 * The binary / operator yields the quotient, and the binary % operator yields the remainder
 * from the division of the first expression by the second.
 * If the second operand of / or % is zero the behavior is undefined; otherwise (a/b)*b + a%b is equal to a.
 * If both operands are nonnegative then the remainder is nonnegative;
 * if not, the sign of the remainder is implementation-defined 74).
 */
static inline int fixed_mod(int a, int b)
{
    int rem = a % b;
    while (rem < 0)
        rem += b;

    return rem;
}

#endif