/** 
  **************************************************************************************************
  *  @file   emacif.h
  *  @brief  Low-level ethernet driver built on top of SAMA5 software package interfaces.
  **************************************************************************************************
  *  Copyright (C) 2020 Eremex
  *  All rights reserved.
  *************************************************************************************************/

#include "native/debug.h"
#include "native/ldr.h"
#include "emacd.h"
#include "lwip/netif.h"
#include "lwip/pbuf.h"
#include "string.h"
#include "native/net_driver.h"

static void emac_input(struct netif *netif);
static err_t emac_output(struct netif *netif, struct pbuf *p);

#define RX_BUFFERS  16
#define TX_BUFFERS  8
#define IST_STACK_SIZE 0x400
#define BOARD_MAINOSC 12000000
#define BOARD_MCK ((unsigned long)((BOARD_MAINOSC / 3 / 2) * 66 )) 

struct emac_dma_bufs_t
{
    sEmacTxDescriptor gTxDs[TX_BUFFERS];
    sEmacRxDescriptor gRxDs[RX_BUFFERS];
    fEmacdTransferCallback gTxCbs[TX_BUFFERS];    
    uint8_t pTxBuffer[TX_BUFFERS * EMAC_TX_UNITSIZE];
    uint8_t pRxBuffer[RX_BUFFERS * EMAC_RX_UNITSIZE];
};

lang_static_assert((offsetof(struct emac_dma_bufs_t, pTxBuffer) & 7) == 0);
lang_static_assert((offsetof(struct emac_dma_bufs_t, pRxBuffer) & 7) == 0);

struct emac_private_t
{
    sEmacd gEmacd;
    Macb gMacb;    
    struct netif* netif;
    fx_sem_t sem;
    fx_interrupt_t intr;
    fx_thread_t thread;
    volatile void* emac_vbase;
    uintptr_t dma_region_vbase;
    uintptr_t dma_region_pbase;
    int stk[IST_STACK_SIZE / sizeof(int)];
    uint8_t input_frame[EMAC_TX_UNITSIZE];
    uint8_t output_frame[EMAC_TX_UNITSIZE];    
};

struct emac_params_t
{
    uint32_t emac_pbase;
    uint32_t emac_size;
    uint32_t irq;
    uint32_t ist_prio;    
    uint32_t dma_pbase;
    uint32_t dma_size;
};

/*
 * Called by EMACD_Handler in case when new received frames are available.
 */
static void emac_rx_callback(sEmacd* arg, uint32_t status)
{
    const struct emac_private_t* priv = lang_containing_record(arg, struct emac_private_t, gEmacd);
    
    if (status & EMAC_RSR_REC)
    {
        emac_input(priv->netif);
    }
}

/*
 * Low level initialization and link detection.
 */
static bool emac_init(struct netif *netif, struct emac_params_t* params)
{
    const struct net_driver_t* net_driver = netif->state;
    struct emac_private_t* priv = net_driver->private;
    struct emac_dma_bufs_t* b = (struct emac_dma_bufs_t*) priv->dma_region_vbase;
    sEmacd* pEmacd = &priv->gEmacd;
    Macb* pMacb = &priv->gMacb;

    EMACD_Init(pEmacd, priv->emac_vbase, params->irq, 0, 0);
    
    EMACD_InitTransfer(pEmacd, 
        b->pRxBuffer, b->gRxDs, RX_BUFFERS, 
        b->pTxBuffer, b->gTxDs, b->gTxCbs, TX_BUFFERS
    );

    EMAC_SetAddress(pEmacd->pHw, 0, netif->hwaddr);
    EMACD_SetRxCallback(pEmacd, emac_rx_callback);
    
    MACB_Init(pMacb, pEmacd, /*BOARD_EMAC_PHY_ADDR*/ 1);
    
    const int phy_init_success = MACB_InitPhy(pMacb, BOARD_MCK);

    if (!phy_init_success)
    {
        DPRINTF(DBG_ERROR, "PHY Initialize ERROR!");
        return false;
    }
    
    /* Auto Negotiate, work in RMII mode */
    const int negotiate_success = MACB_AutoNegotiate(pMacb, /* BOARD_EMAC_MODE_RMII */ 1 );

    if (!negotiate_success)
    {
        DPRINTF(DBG_ERROR, "Auto Negotiate ERROR!") ;
        return false;
    }

    return true;
}

/*
 * Frame transmission.
 * TODO: Implement packet queuing. There may be synchronization bugs with current design.
 */
static err_t emac_output(struct netif *netif, struct pbuf *p)
{
    const struct net_driver_t* d = netif->state;
    struct emac_private_t* priv = d->private;
    uint8_t* bufptr = priv->output_frame;
    struct pbuf *q;

    DPRINTF(DBG_INFO, "Output frame 0x%x bytes", p->tot_len);     

    for(q = p; q != NULL; q = q->next) {   
        memcpy(bufptr, q->payload, q->len);
        bufptr += q->len;
    }

    const uint8_t rc = EMACD_Send(&priv->gEmacd, priv->output_frame, p->tot_len, NULL);

    DPRINTF(DBG_INFO, "Output status 0x%x", rc);  

    if (rc != EMACD_OK) {
        return ERR_BUF;
    }

    return ERR_OK;
}

/*
 * Frame transmission.
 */
static void emac_input(struct netif *netif)
{
    const struct net_driver_t* d = netif->state;
    struct emac_private_t* p = d->private;
    uint8_t* const bufptr = p->input_frame;
    uint32_t frmlen;

    EMACD_Poll(&p->gEmacd, bufptr, sizeof(p->input_frame), &frmlen);

    DPRINTF(DBG_INFO, "Received length 0x%x", frmlen);
    
    if (frmlen > 0)
    {
        d->input(netif, bufptr, frmlen);
    }
}

/*
 * Interrupt service thread.
 */
static void emac_intr_thread(void* arg)
{
    struct emac_private_t* priv = arg;

    for (;;) {
        const int error = fx_sem_timedwait(&priv->sem, FX_THREAD_INFINITE_TIMEOUT);
        
        fx_interrupt_mask(&priv->intr, false);

        DPRINTF(DBG_INFO, "EMAC IST");

        if (error)
        {
            fx_thread_exit();
        }

        EMACD_Handler(&priv->gEmacd);
    }
}

void* emac_addr_p2v(sEmacd* arg, uintptr_t paddr)
{
    const struct emac_private_t* priv = lang_containing_record(arg, struct emac_private_t, gEmacd);
    return (void*) (paddr - priv->dma_region_pbase + priv->dma_region_vbase);
}

void* emac_addr_v2p(sEmacd* arg, void* addr)
{
    const struct emac_private_t* priv = lang_containing_record(arg, struct emac_private_t, gEmacd);
    return (void*) ((uintptr_t)addr - priv->dma_region_vbase + priv->dma_region_pbase);
}

bool emac_read_cfg_key(const char* devname, const char* keyname, uint32_t* value)
{
    const char root[] = "/devices/";
    const size_t device_name_len = strlen(devname);
    const size_t key_len = strlen(keyname);
    const size_t path_len = sizeof(root) + device_name_len + key_len + 1;
    char key[path_len];

    memset(key, 0, path_len);
    strcat(key, root);
    strcat(key, devname);
    strcat(key, keyname);

    return fx_reg_get_value(key, FX_REG_TYPE_DWORD, value, sizeof(uint32_t)) == 0;
}

bool emac_read_cfg_params(char* emac_devname, struct emac_params_t* param)
{
    const char root[] = "/devices/";
    const char dma_region_key[] = "/dma_region";
    const size_t emac_name_len = strlen(emac_devname);  
    const size_t path_len = sizeof(root) + emac_name_len + sizeof(dma_region_key) + 1;  
    char key[path_len];
    char dma_region_name[20] = "";

    if (!emac_read_cfg_key(emac_devname, "/base", &param->emac_pbase)) {
        return false;
    }

    if (!emac_read_cfg_key(emac_devname, "/size", &param->emac_size)) {
        return false;
    }

    if (!emac_read_cfg_key(emac_devname, "/irq", &param->irq)) {
        return false;
    }

    if (!emac_read_cfg_key(emac_devname, "/ist_prio", &param->ist_prio)) {
        return false;
    }

    memset(key, 0, path_len);
    strcat(key, root);
    strcat(key, emac_devname);
    strcat(key, dma_region_key);

    if (fx_reg_get_value(key, FX_REG_TYPE_STRING, dma_region_name, sizeof(dma_region_name))) {
        return false;
    }
    
    if (!emac_read_cfg_key(dma_region_name, "/base", &param->dma_pbase)) {
        return false;
    }

    if (!emac_read_cfg_key(dma_region_name, "/size", &param->dma_size)) {
        return false;
    }

    return true;
}

/*
 * EMAC init function called by link layer driver.
 */
err_t ethernet_init(struct netif *netif)
{
    static struct emac_private_t g_priv;
    struct emac_private_t* priv = &g_priv;
    struct net_driver_t* d = netif->state;
    struct emac_params_t params;

    if (!emac_read_cfg_params(d->private, &params)) {
        return ERR_MEM;
    }

    DPRINTF(DBG_INFO, "EMAC base=0x%x size=0x%x IRQ=0x%x IST prio=0x%x" , 
        params.emac_pbase, 
        params.emac_size,
        params.irq,
        params.ist_prio);

    DPRINTF(DBG_INFO, "EMAC DMA base=0x%x size=0x%x", params.dma_pbase, params.dma_size);

    d->private = priv;
    d->output = emac_output;

    netif->hwaddr_len = 6;
    netif->hwaddr[0] = 0x3a;
    netif->hwaddr[1] = 0x1f;
    netif->hwaddr[2] = 0x34;
    netif->hwaddr[3] = 0x08;
    netif->hwaddr[4] = 0x54;
    netif->hwaddr[5] = 0x54;
    netif->mtu = 1500;

    priv->netif = netif;
    priv->emac_vbase = fx_dev_map(params.emac_pbase, params.emac_size); 

    if (priv->emac_vbase == NULL) {
        DPRINTF(DBG_ERROR, "EMAC device at addr 0x%x not mapped", params.emac_pbase);
        return ERR_MEM;
    }

    priv->dma_region_pbase = params.dma_pbase;
    priv->dma_region_vbase = (uintptr_t) fx_dev_map(params.dma_pbase, params.dma_size);
    
    if (priv->dma_region_vbase == 0 || params.dma_size < sizeof(struct emac_dma_bufs_t)) {
        DPRINTF(DBG_ERROR, "EMAC DMA memory at addr 0x%x not mapped", params.dma_pbase);
        return ERR_MEM;
    }  

    memset((void*) priv->dma_region_vbase, 0, sizeof(struct emac_dma_bufs_t));

    if (!emac_init(netif, &params)) {
        return ERR_MEM;        
    }

    int error = fx_sem_init(&priv->sem, 0, 256, FX_SYNC_POLICY_FIFO);

    if (error) {
        DPRINTF(DBG_ERROR, "Can't create semaphore", error);
        return ERR_MEM;
    }

    error = fx_interrupt_init(&priv->intr, &priv->sem, params.irq);

    if (error) {
        DPRINTF(DBG_ERROR, "Can't create interrupt object", error);
        goto sem_deinit;
    }

    error = fx_interrupt_attach(&priv->intr);

    if (error) {
        DPRINTF(DBG_ERROR, "Can't attach interrupt", error);
        goto intr_deinit;
    }

    error = fx_thread_init(
        &priv->thread, emac_intr_thread, priv, 
        params.ist_prio, priv->stk, sizeof(priv->stk), false
    );

    if (error) {
        DPRINTF(DBG_ERROR, "Can't create IST", error);
        goto intr_deinit;
    }
    
    netif->flags = NETIF_FLAG_BROADCAST | NETIF_FLAG_ETHARP | NETIF_FLAG_LINK_UP;
    return ERR_OK;

intr_deinit:
    fx_interrupt_deinit(&priv->intr);
sem_deinit:
    fx_sem_deinit(&priv->sem);

    return ERR_MEM;
}
