/**
**************************************************************************************************
* @file   ethernetif.c
* @brief  Interface of Ethernet driver for LPC17xx to LwIP
**************************************************************************************************
* Copyright(C) 2020 Eremex
* All rights reserved.
* ************************************************************************************************/

/*
 * lwIP includes.
 */
#include "lwip/sys.h"
#include <lwip/stats.h>
#include <lwip/snmp.h>
#include "netif/etharp.h"

#include "lpc17xx_eth.h"
#include <FXRTOS.h>
#include <string.h>

#define ETH_PAD_SIZE 0

/*
 * Device extension.
 */
struct drv_eth_extension
{
  lpc17xx_eth_dev_t dev;

  /* The network interface that was opened. */
  struct netif *lwip_netif;

  /* Semaphore to be posted on Ethernet events. */
  fx_sem_t intr_sem;

  /* Interrupt service thread and stack. */
  fx_thread_t intr_thread;
  int intr_thread_stk[0x400];
};

/*-----------------------------------------------------------*/

/**
 * This function should do the actual transmission of the packet. The packet is
 * contained in the pbuf that is passed to the function. This pbuf
 * might be chained.
 */

static err_t 
drv_eth_output( struct netif *drv_netif, struct pbuf *p )
{
  unsigned char* buf;
  u16_t packet_len = p->tot_len - ETH_PAD_SIZE;
  size_t buf_sz = packet_len;
  err_t error = ERR_OK;
  struct drv_eth_extension* dev_ext = (struct drv_eth_extension*) drv_netif->state;

  lpc17xx_eth_acquire_tx(&dev_ext->dev, (void**) &buf, &buf_sz);

  /* Initiate transfer. */
  if( p->len == p->tot_len ) 
  {
	  memcpy(buf, &( ( unsigned char * ) p->payload )[ ETH_PAD_SIZE ], p->len);
  } 
  else 
  {
    /* pbuf chain, copy into contiguous buf. */
    if( p->tot_len >= 1522 ) 
    {
      LINK_STATS_INC( link.lenerr );
      LINK_STATS_INC( link.drop );
      snmp_inc_ifoutdiscards( drv_netif );
      error = ERR_BUF;
    }
    else
    {
      struct pbuf *q;
      unsigned char* buf_ptr = buf;

      for( q = p; q != NULL; q = q->next ) 
      {
        /* Send the data from the pbuf to the interface, one pbuf at a
          time. The size of the data in each pbuf is kept in the ->len
          variable. */
        /* send data from(q->payload, q->len); */
        LWIP_DEBUGF( NETIF_DEBUG, ("NETIF: send buf_ptr %p q->payload %p q->len %i q->next %p\n", buf_ptr, q->payload, ( int ) q->len, ( void* ) q->next ) );
        if( q == p ) 
        {
          memcpy( buf_ptr, &( ( char * ) q->payload )[ ETH_PAD_SIZE ], q->len - ETH_PAD_SIZE );
          buf_ptr += q->len - ETH_PAD_SIZE;
        } 
        else 
        {
          memcpy( buf_ptr, q->payload, q->len );
          buf_ptr += q->len;
        }
      }
    }
  }

  if( error == ERR_OK )
  {
    /* signal that packet should be sent */
    if( !lpc17xx_eth_send( &dev_ext->dev, buf, packet_len) ) 
    {
      LINK_STATS_INC( link.memerr );
      LINK_STATS_INC( link.drop );
      snmp_inc_ifoutdiscards( drv_netif );
      error = ERR_BUF;
    }
    else
    {
      struct eth_hdr *hdr;
      
      LINK_STATS_INC( link.xmit );
      snmp_add_ifoutoctets( drv_netif, packet_len );
      hdr = ( struct eth_hdr * )p->payload;

      if( ( hdr->dest.addr[ 0 ] & 1 ) != 0 ) 
      {
	/* broadcast or multicast packet*/
 	snmp_inc_ifoutnucastpkts( drv_netif );
      } 
      else 
      {
	/* unicast packet */
	snmp_inc_ifoutucastpkts( drv_netif );
      }
    }
  }
  
  return error;
}

/*-----------------------------------------------------------*/

/**
 * Should allocate a pbuf and transfer the bytes of the incoming
 * packet from the interface into the pbuf.
 */
static struct pbuf*
drv_eth_input_helper( const unsigned char* const input_data, long input_len )
{
  struct pbuf *p = NULL, *q;

	if( input_len > 0 )
	{
		#if ETH_PAD_SIZE
			len += ETH_PAD_SIZE; /* allow room for Ethernet padding */
		#endif

		/* We allocate a pbuf chain of pbufs from the pool. */
		p = pbuf_alloc( PBUF_RAW, input_len, PBUF_POOL );
  
		if( p != NULL ) 
		{
			#if ETH_PAD_SIZE
				pbuf_header( p, -ETH_PAD_SIZE ); /* drop the padding word */
			#endif

			/* We iterate over the pbuf chain until we have read the entire
			* packet into the pbuf. */
			input_len = 0;
			for( q = p; q != NULL; q = q->next ) 
			{
				/* Read enough bytes to fill this pbuf in the chain. The
				* available data in the pbuf is given by the q->len
				* variable.
				* This does not necessarily have to be a memcpy, you can also preallocate
				* pbufs for a DMA-enabled MAC and after receiving truncate it to the
				* actually received size. In this case, ensure the packet_len member of the
				* pbuf is the sum of the chained pbuf len members.
				*/
				memcpy( q->payload, &( input_data[ input_len ] ), q->len );
				input_len += q->len;
			}

			#if ETH_PAD_SIZE
				pbuf_header( p, ETH_PAD_SIZE ); /* reclaim the padding word */
			#endif

			LINK_STATS_INC( link.recv );
		}
	}

	return p;  
}

/*-----------------------------------------------------------*/

/**
 * This function should be called when a packet is ready to be read
 * from the interface. It uses the function drv_eth_input_helper() that
 * should handle the actual reception of bytes from the network
 * interface. Then the type of the received packet is determined and
 * the appropriate input function is called.
 */
static void 
drv_eth_input(struct netif *drv_netif)
{
  struct eth_hdr* hdr;
  struct pbuf *p;
  unsigned char* input_data;
  size_t input_len;
  struct drv_eth_extension* dev_ext = (struct drv_eth_extension*) drv_netif->state;

  lpc17xx_eth_receive(&dev_ext->dev, (void**) &input_data, &input_len);

  /* move received packet into a new pbuf */
  p = drv_eth_input_helper( input_data, input_len );

  /* no packet could be read, silently ignore this */
  if( p != NULL ) 
  {
	/* points to packet payload, which starts with an Ethernet header */
	hdr = p->payload;

	switch( htons( hdr->type ) ) 
	{
		/* IP or ARP packet? */
		case ETHTYPE_IP:
		case ETHTYPE_ARP:

			/* full packet send to tcpip_thread to process */
			if( drv_netif->input( p, drv_netif ) != ERR_OK )
			{ 
				LWIP_DEBUGF(NETIF_DEBUG, ( "ethernetif_input: IP input error\n" ) );
				pbuf_free(p);
				p = NULL;
			}
			break;

			default:
				pbuf_free( p );
				p = NULL;
			break;
		}
  }

  lpc17xx_eth_release_rx(&dev_ext->dev, (void*) input_data, input_len);
}

/*-----------------------------------------------------------*/

static void 
drv_eth_input_signal(struct netif *drv_netif)
{
  struct drv_eth_extension* dev_ext = (struct drv_eth_extension*) drv_netif->state;
  fx_sem_post(&dev_ext->intr_sem);
}


/*-----------------------------------------------------------*/

/* Interrupt service thread. */
static void intr_thread(void* arg)
{
  struct drv_eth_extension* dev_ext = (struct drv_eth_extension*) arg;
  
  while(1)
  {
	/* Wait for signal from privileged part of application. */
	fx_sem_timedwait(&dev_ext->intr_sem, FX_THREAD_INFINITE_TIMEOUT);

	/* Handle request. */
	drv_eth_input(dev_ext->lwip_netif);
  }
}

/*-----------------------------------------------------------*/

static struct drv_eth_extension device_extension ;

/**
 * Should be called at the beginning of the program to set up the
 * network interface. It calls the function drv_eth_init_hardware() to do the
 * actual setup of the hardware.
 *
 * This function should be passed as a parameter to netif_add().
 */
err_t 
drv_eth_init( struct netif *drv_netif )
{
  err_t res = ERR_OK;
  struct drv_eth_extension* dev_ext = &device_extension;

  LWIP_ASSERT( "drv_netif != NULL", ( drv_netif != NULL ) );
	
  #if LWIP_NETIF_HOSTNAME
  {
	/* Initialize interface hostname */
	drv_netif->hostname = "lwip";
  }
  #endif /* LWIP_NETIF_HOSTNAME */

  memset(&device_extension, 0, sizeof(device_extension));

  drv_netif->state = dev_ext;
  drv_netif->name[ 0 ] = 'a';
  drv_netif->name[ 1 ] = 'b';

  /* We directly use etharp_output() here to save a function call.
  * You can instead declare your own function an call etharp_output()
  * from it if you have to do some checks before sending (e.g. if link
  * is available...) */
  drv_netif->output = etharp_output;
  drv_netif->flags = NETIF_FLAG_BROADCAST | NETIF_FLAG_ETHARP | NETIF_FLAG_LINK_UP;
  drv_netif->hwaddr_len = 6;
  drv_netif->mtu = 1522;
  drv_netif->linkoutput = drv_eth_output;

  dev_ext->lwip_netif = drv_netif;

  lpc17xx_eth_init(&dev_ext->dev, (void (*)(void*)) drv_eth_input_signal, drv_netif, NULL);
  lpc17xx_eth_get_mac_addr(&dev_ext->dev, drv_netif->hwaddr);

  fx_sem_init(&dev_ext->intr_sem, 0, 255, FX_SYNC_POLICY_DEFAULT);

  fx_thread_init(
    &dev_ext->intr_thread, 
    intr_thread, dev_ext, 10, 
    (void*)(dev_ext->intr_thread_stk), sizeof(dev_ext->intr_thread_stk), 
    false);

  return res;
}

void
drv_eth_interrupt(void)
{
  struct drv_eth_extension* dev_ext = &device_extension;
  lpc17xx_eth_isr_handler(&dev_ext->dev);
}
