/** 
  **************************************************************************************************
  *  @file   lpc17xx_eth.c
  *  @brief  Implementation of Ethernet driver for LPC17xx boards.
  **************************************************************************************************
  *  Copyright (C) 2020 Eremex
  *  All rights reserved.
  *************************************************************************************************/

#include "lpc17xx_eth.h"
#include "LPC17xx.h"

/* MAC Configuration Register 1 */
#define MAC1_REC_EN         0x00000001  /* Receive Enable                    */
#define MAC1_PASS_ALL       0x00000002  /* Pass All Receive Frames           */
#define MAC1_RX_FLOWC       0x00000004  /* RX Flow Control                   */
#define MAC1_TX_FLOWC       0x00000008  /* TX Flow Control                   */
#define MAC1_LOOPB          0x00000010  /* Loop Back Mode                    */
#define MAC1_RES_TX         0x00000100  /* Reset TX Logic                    */
#define MAC1_RES_MCS_TX     0x00000200  /* Reset MAC TX Control Sublayer     */
#define MAC1_RES_RX         0x00000400  /* Reset RX Logic                    */
#define MAC1_RES_MCS_RX     0x00000800  /* Reset MAC RX Control Sublayer     */
#define MAC1_SIM_RES        0x00004000  /* Simulation Reset                  */
#define MAC1_SOFT_RES       0x00008000  /* Soft Reset MAC                    */

/* MAC Configuration Register 2 */
#define MAC2_FULL_DUP       0x00000001  /* Full Duplex Mode                  */
#define MAC2_FRM_LEN_CHK    0x00000002  /* Frame Length Checking             */
#define MAC2_HUGE_FRM_EN    0x00000004  /* Huge Frame Enable                 */
#define MAC2_DLY_CRC        0x00000008  /* Delayed CRC Mode                  */
#define MAC2_CRC_EN         0x00000010  /* Append CRC to every Frame         */
#define MAC2_PAD_EN         0x00000020  /* Pad all Short Frames              */
#define MAC2_VLAN_PAD_EN    0x00000040  /* VLAN Pad Enable                   */
#define MAC2_ADET_PAD_EN    0x00000080  /* Auto Detect Pad Enable            */
#define MAC2_PPREAM_ENF     0x00000100  /* Pure Preamble Enforcement         */
#define MAC2_LPREAM_ENF     0x00000200  /* Long Preamble Enforcement         */
#define MAC2_NO_BACKOFF     0x00001000  /* No Backoff Algorithm              */
#define MAC2_BACK_PRESSURE  0x00002000  /* Backoff Presurre / No Backoff     */
#define MAC2_EXCESS_DEF     0x00004000  /* Excess Defer                      */

/* Test Register */
#define TEST_SHCUT_PQUANTA  0x00000001  /* Shortcut Pause Quanta             */
#define TEST_TST_PAUSE      0x00000002  /* Test Pause                        */
#define TEST_TST_BACKP      0x00000004  /* Test Back Pressure                */

/* MII Management Configuration Register */
#define MCFG_SCAN_INC       0x00000001  /* Scan Increment PHY Address        */
#define MCFG_SUPP_PREAM     0x00000002  /* Suppress Preamble                 */
#define MCFG_CLK_SEL        0x0000001C  /* Clock Select Mask                 */
#define MCFG_RES_MII        0x00008000  /* Reset MII Management Hardware     */

/* Clock Host dividers                  */
#define HOST_CLK_BY_4   (0x00 << 2) /* Host Clock divided by 4           */
#define HOST_CLK_BY_6   (0x02 << 2) /* Host Clock divided by 6           */
#define HOST_CLK_BY_8   (0x03 << 2) /* Host Clock divided by 8           */
#define HOST_CLK_BY_10    (0x04 << 2) /* Host Clock divided by 10          */
#define HOST_CLK_BY_14    (0x05 << 2) /* Host Clock divided by 14          */
#define HOST_CLK_BY_20    (0x06 << 2) /* Host Clock divided by 20          */
#define HOST_CLK_BY_28    (0x07 << 2) /* Host Clock divided by 28          */

/* MII Management Command Register */
#define MCMD_READ           0x00000001  /* MII Read                          */
#define MCMD_SCAN           0x00000002  /* MII Scan continuously             */

#define MII_WR_TOUT         0x00050000  /* MII Write timeout count           */
#define MII_RD_TOUT         0x00050000  /* MII Read timeout count            */

/* MII Management Address Register */
#define MADR_REG_ADR        0x0000001F  /* MII Register Address Mask         */
#define MADR_PHY_ADR        0x00001F00  /* PHY Address Mask                  */

/* MII Management Indicators Register */
#define MIND_BUSY           0x00000001  /* MII is Busy                       */
#define MIND_SCAN           0x00000002  /* MII Scanning in Progress          */
#define MIND_NOT_VAL        0x00000004  /* MII Read Data not valid           */
#define MIND_MII_LINK_FAIL  0x00000008  /* MII Link Failed                   */

/* Command Register */
#define CR_RX_EN            0x00000001  /* Enable Receive                    */
#define CR_TX_EN            0x00000002  /* Enable Transmit                   */
#define CR_REG_RES          0x00000008  /* Reset Host Registers              */
#define CR_TX_RES           0x00000010  /* Reset Transmit Datapath           */
#define CR_RX_RES           0x00000020  /* Reset Receive Datapath            */
#define CR_PASS_RUNT_FRM    0x00000040  /* Pass Runt Frames                  */
#define CR_PASS_RX_FILT     0x00000080  /* Pass RX Filter                    */
#define CR_TX_FLOW_CTRL     0x00000100  /* TX Flow Control                   */
#define CR_RMII             0x00000200  /* Reduced MII Interface             */
#define CR_FULL_DUP         0x00000400  /* Full Duplex                       */

/* Non Back-to-Back Inter-Packet-Gap Register */
#define IPGR_DEF            0x00000012  /* Recommended value                 */

/* Collision Window/Retry Register */
#define CLRT_DEF            0x0000370F  /* Default value                     */

/* PHY Support Register */
#define SUPP_SPEED          0x00000100  /* Reduced MII Logic Current Speed   */

//----------------------------------------------------------------------------------------------------

#define DEFAULT_DESCR_NUM 3
#define DEFAULT_MAX_FRAME_LEN 1532
#define DEFAULT_DESCR_MEM_AREA 0x2007C000
#define DEFAULT_DESCR_MEM_SZ (DEFAULT_DESCR_NUM*8*3 + DEFAULT_DESCR_NUM*4 + DEFAULT_MAX_FRAME_LEN*DEFAULT_DESCR_NUM*2)

//----------------------------------------------------------------------------------------------------

#define PHY_ADDRESS 1

static void 
lpc17xx_eth_write_PHY(uint32_t reg, uint32_t val) 
{
  unsigned i;

  LPC_EMAC->MADR = (PHY_ADDRESS << 8) | reg;
  LPC_EMAC->MWTD = val;

  for(i = 0; (i < 500) && (LPC_EMAC->MIND & MIND_BUSY); ++i) fx_thread_sleep(1);
}

//----------------------------------------------------------------------------------------------------

static uint16_t
lpc17xx_eth_read_PHY(unsigned char reg) 
{
  unsigned i;

  LPC_EMAC->MADR = (PHY_ADDRESS << 8) | reg;
  LPC_EMAC->MCMD = MCMD_READ;

  for(i = 0; (i < 500) && (LPC_EMAC->MIND & MIND_BUSY); ++i) fx_thread_sleep(1);
  LPC_EMAC->MCMD = 0;

  return (uint16_t) LPC_EMAC->MRDD;
}

//----------------------------------------------------------------------------------------------------

/* DP83848C PHY Registers */
#define PHY_REG_BMCR        0x00        /* Basic Mode Control Register       */
#define PHY_REG_BMSR        0x01        /* Basic Mode Status Register        */
#define PHY_REG_IDR1        0x02        /* PHY Identifier 1                  */
#define PHY_REG_IDR2        0x03        /* PHY Identifier 2                  */
#define PHY_REG_ANAR        0x04        /* Auto-Negotiation Advertisement    */
#define PHY_REG_ANLPAR      0x05        /* Auto-Neg. Link Partner Abitily    */
#define PHY_REG_ANER        0x06        /* Auto-Neg. Expansion Register      */
#define PHY_REG_ANNPTR      0x07        /* Auto-Neg. Next Page TX            */
#define PHY_FULLD_100M      0x2100      /* Full Duplex 100Mbit               */
#define PHY_HALFD_100M      0x2000      /* Half Duplex 100Mbit               */
#define PHY_FULLD_10M       0x0100      /* Full Duplex 10Mbit                */
#define PHY_HALFD_10M       0x0000      /* Half Duplex 10MBit                */
#define PHY_AUTO_NEG        0x3000      /* Select Auto Negotiation           */

/* PHY Extended Registers */
#define PHY_REG_STS         0x10        /* Status Register                   */
#define PHY_REG_MICR        0x11        /* MII Interrupt Control Register    */
#define PHY_REG_MISR        0x12        /* MII Interrupt Status Register     */
#define PHY_REG_FCSCR       0x14        /* False Carrier Sense Counter       */
#define PHY_REG_RECR        0x15        /* Receive Error Counter             */
#define PHY_REG_PCSR        0x16        /* PCS Sublayer Config. and Status   */
#define PHY_REG_RBR         0x17        /* RMII and Bypass Register          */
#define PHY_REG_LEDCR       0x18        /* LED Direct Control Register       */
#define PHY_REG_PHYCR       0x19        /* PHY Control Register              */
#define PHY_REG_10BTSCR     0x1A        /* 10Base-T Status/Control Register  */
#define PHY_REG_CDCTRL1     0x1B        /* CD Test Control and BIST Extens.  */
#define PHY_REG_EDCR        0x1D        /* Energy Detect Control Register    */

#define DP83848C_ID         0x20005C90  /* National PHY Identifier           */

#define PHY_BMCR_SOFT_RESET   0x8000
#define PHY_AUTO_NEG_COMPLETE 0x0020
#define PHY_LINK_STATUS_ON    0x0001

//----------------------------------------------------------------------------------------------------

static bool 
lpc17xx_eth_national_phy_init(bool* full_duplex, bool* full_speed)
{
  unsigned i;
  uint32_t regv, id1, id2;
  bool error;

  /* Check if this is a DP83848C PHY. */
  id1 = lpc17xx_eth_read_PHY(PHY_REG_IDR1);
  id2 = lpc17xx_eth_read_PHY(PHY_REG_IDR2);

  if(((id1 << 16) | (id2 & 0xFFF0)) != DP83848C_ID) 
  {
    /* Not DP83848C chip. Abort. */ 
    return false;
  }

  /* Put the PHY chip in reset mode. */
  lpc17xx_eth_write_PHY(PHY_REG_BMCR, PHY_BMCR_SOFT_RESET);

  error = true;

  /* Wait for hardware reset to end. */
  for (i = 0; i < 0x100; ++i) 
  {
    regv = lpc17xx_eth_read_PHY(PHY_REG_BMCR);

    if (!(regv & PHY_BMCR_SOFT_RESET)) 
    {
      /* Reset complete. */
      error = false;
      break;
    }

    fx_thread_sleep(1);
  }

  if(error)
  {
    /* Cannot reset PHY chip. Abort. */
    return false;
  }

  /* Configure the PHY device */

  /* Use autonegotiation about the link speed. */
  lpc17xx_eth_write_PHY(PHY_REG_BMCR, PHY_AUTO_NEG);

  error = true;

  /* Wait to complete Auto_Negotiation. */
  for(i = 0; i < 0x500; ++i) 
  {
    regv = lpc17xx_eth_read_PHY(PHY_REG_BMSR);

    if (regv & PHY_AUTO_NEG_COMPLETE) 
    {
      /* Autonegotiation Complete. */
      error = false;
      break;
    }

    fx_thread_sleep(1);
  }

  if(error)
  {
    /* Negotiation failed. */
    return false;
  }
  
  error = true;

  /* Check the link status. */
  for (i = 0; i < 0x500; ++i) 
  {
    regv = lpc17xx_eth_read_PHY(PHY_REG_STS);

    if (regv & PHY_LINK_STATUS_ON) 
    {
      /* Link is on. */
      error = false;
      break;
    }
  }

  if(error)
  {
    /* Link is off failed. */
    return false;
  }

  regv = lpc17xx_eth_read_PHY(PHY_REG_STS);

  *full_duplex = (regv & 0x0004) != 0;
  *full_speed = (regv & 0x0002) == 0;

  return true;
}

//----------------------------------------------------------------------------------------------------

/* RX Descriptor Control Word */
#define RCTRL_SIZE          0x000007FF  /* Buffer size mask                  */
#define RCTRL_INT           0x80000000  /* Generate RxDone Interrupt         */

/* TX Descriptor Control Word */
#define TCTRL_SIZE          0x000007FF  /* Size of data buffer in bytes      */
#define TCTRL_OVERRIDE      0x04000000  /* Override Default MAC Registers    */
#define TCTRL_HUGE          0x08000000  /* Enable Huge Frame                 */
#define TCTRL_PAD           0x10000000  /* Pad short Frames to 64 bytes      */
#define TCTRL_CRC           0x20000000  /* Append a hardware CRC to Frame    */
#define TCTRL_LAST          0x40000000  /* Last Descriptor for TX Frame      */
#define TCTRL_INT           0x80000000  /* Generate TxDone Interrupt         */

static void 
lpc17xx_eth_init_descriptors(lpc17xx_eth_dev_t* dev, void* frame_area, size_t frame_area_sz)
{
  unsigned i;
  uint8_t* const rx_buf = (uint8_t*) frame_area;
  uint8_t* const tx_buf = (uint8_t*) frame_area + (dev->rx_descriptors_num * dev->max_frame_len);

  /* Initialize Rx DMA Descriptors */
  for (i = 0; i < dev->rx_descriptors_num; ++i) 
  {
    dev->rx_descriptors[i].buffer = (uint32_t) (rx_buf + i * dev->max_frame_len);
    dev->rx_descriptors[i].control = RCTRL_INT | (dev->max_frame_len-1);
    dev->rx_status[i].status = dev->rx_status[i].status_hash_crc = 0;
  }  

  /* Initialize Tx DMA Descriptors */
  for (i = 0; i < dev->tx_descriptors_num; ++i) 
  {
    dev->tx_descriptors[i].buffer = (uint32_t) (tx_buf + i * dev->max_frame_len);
    dev->tx_status[i] = dev->tx_descriptors[i].control = 0;
  }  
}

//----------------------------------------------------------------------------------------------------

/* Back-to-Back Inter-Packet-Gap Register */
#define IPGT_FULL_DUP       0x00000015  /* Recommended value for Full Duplex */
#define IPGT_HALF_DUP       0x00000012  /* Recommended value for Half Duplex */

/* Receive Filter Control Register */
#define RFC_UCAST_EN        0x00000001  /* Accept Unicast Frames Enable      */
#define RFC_BCAST_EN        0x00000002  /* Accept Broadcast Frames Enable    */
#define RFC_MCAST_EN        0x00000004  /* Accept Multicast Frames Enable    */
#define RFC_UCAST_HASH_EN   0x00000008  /* Accept Unicast Hash Filter Frames */
#define RFC_MCAST_HASH_EN   0x00000010  /* Accept Multicast Hash Filter Fram.*/
#define RFC_PERFECT_EN      0x00000020  /* Accept Perfect Match Enable       */
#define RFC_MAGP_WOL_EN     0x00001000  /* Magic Packet Filter WoL Enable    */
#define RFC_PFILT_WOL_EN    0x00002000  /* Perfect Filter WoL Enable         */

/* Interrupt Status/Enable/Clear/Set Registers */
#define INT_RX_OVERRUN      0x00000001  /* Overrun Error in RX Queue         */
#define INT_RX_ERR          0x00000002  /* Receive Error                     */
#define INT_RX_FIN          0x00000004  /* RX Finished Process Descriptors   */
#define INT_RX_DONE         0x00000008  /* Receive Done                      */
#define INT_TX_UNDERRUN     0x00000010  /* Transmit Underrun                 */
#define INT_TX_ERR          0x00000020  /* Transmit Error                    */
#define INT_TX_FIN          0x00000040  /* TX Finished Process Descriptors   */
#define INT_TX_DONE         0x00000080  /* Transmit Done                     */
#define INT_SOFT_INT        0x00001000  /* Software Triggered Interrupt      */
#define INT_WAKEUP          0x00002000  /* Wakeup Event Interrupt            */

#define MAC_6         0x10       /* MAC_6 */
#define MAC_5         0x1f       /* MAC_5 */
#define MAC_4         0xe0       /* MAC_4 */
#define MAC_3         0x12       /* MAC_3 */
#define MAC_2         0x1d       /* MAC_2 */
#define MAC_1         0x0c       /* MAC_1 */

bool 
lpc17xx_eth_init(lpc17xx_eth_dev_t* dev, void (*rx_signal)(void*), void* rx_arg, lpc17xx_eth_params_t* info)
{
  unsigned i;
  bool full_duplex;
  bool full_speed;
  size_t offset;

  lpc17xx_eth_params_t default_params;
  lpc17xx_eth_params_t* const params = info ? info : &default_params;

  if(!dev || !rx_signal)
  {
    return false;
  }

  default_params.phy_init = lpc17xx_eth_national_phy_init;
  default_params.tx_descr_num = default_params.rx_descr_num = DEFAULT_DESCR_NUM;
  default_params.max_frame_len = DEFAULT_MAX_FRAME_LEN;
  default_params.mem = DEFAULT_DESCR_MEM_AREA;
  default_params.mem_sz = DEFAULT_DESCR_MEM_SZ;
  default_params.zero_copy = false;
  default_params.mac_addr[0] = MAC_1; 
  default_params.mac_addr[1] = MAC_2;
  default_params.mac_addr[2] = MAC_3;
  default_params.mac_addr[3] = MAC_4;
  default_params.mac_addr[4] = MAC_5;
  default_params.mac_addr[5] = MAC_6;

  dev->max_frame_len = params->max_frame_len;
  dev->mem_area = (uint8_t*) params->mem;
  dev->mem_sz = params->mem_sz;
  dev->tx_descriptors_num = params->tx_descr_num;
  dev->rx_descriptors_num = params->rx_descr_num;
  dev->rx_signal = rx_signal;
  dev->rx_arg = rx_arg;
  dev->rx_service_req = 0;
  dev->tx_consumed = 0;

  /* Power Up the EMAC controller. */
  LPC_SC->PCONP |= 0x40000000;

  /* Configure Ethernet pins. */
  LPC_PINCON->PINSEL2 = 0x50150105;

  /* Configure pins for RMII and MIIM. */
  LPC_PINCON->PINSEL3 = (LPC_PINCON->PINSEL3 & ~0x0000000F) | 0x00000005;
  
  /* Reset all EMAC internal modules. */
  LPC_EMAC->MAC1= MAC1_RES_TX | MAC1_RES_MCS_TX | MAC1_RES_RX | MAC1_RES_MCS_RX | MAC1_SIM_RES | MAC1_SOFT_RES;
  LPC_EMAC->Command = CR_REG_RES | CR_TX_RES | CR_RX_RES;

  /* A short delay after reset. */
  for (i = 100; i; --i);

  /* Initialize MAC control registers. */
  LPC_EMAC->MAC1 = MAC1_PASS_ALL;
  LPC_EMAC->MAC2 = MAC2_CRC_EN | MAC2_PAD_EN;
  LPC_EMAC->MAXF = params->max_frame_len;
  LPC_EMAC->CLRT = CLRT_DEF;
  LPC_EMAC->IPGR = IPGR_DEF;

  /* Enable Reduced MII interface. */
  LPC_EMAC->Command = CR_RMII | CR_PASS_RUNT_FRM;

  /* MII Mgmt Configuration register and MII Mgnt hardware Reset       */
  /* host clock divided by 20, no suppress preamble, no scan increment */
  LPC_EMAC->MCFG = HOST_CLK_BY_20 | MCFG_RES_MII; 
  for ( i = 0; i < 0x400; ++i );
  LPC_EMAC->MCFG &= (~MCFG_RES_MII);  /* Clear the reset */
  LPC_EMAC->MCMD = 0; 

  if(!params->phy_init(&full_duplex, &full_speed))
  {
    return false;
  }

  /* Configure Full/Half Duplex mode. */
  if (full_duplex) 
  {
    /* Full duplex is enabled. */
    LPC_EMAC->MAC2    |= MAC2_FULL_DUP;
    LPC_EMAC->Command |= CR_FULL_DUP;
    LPC_EMAC->IPGT     = IPGT_FULL_DUP;
  }
  else 
  {
    /* Half duplex mode. */
    LPC_EMAC->IPGT = IPGT_HALF_DUP;
  }

  /* Configure 100MBit/10MBit mode. */
  LPC_EMAC->SUPP = full_speed ? SUPP_SPEED : 0;

  /* Set the Ethernet MAC Address registers */
  LPC_EMAC->SA0 = (params->mac_addr[5] << 8) | params->mac_addr[4];
  LPC_EMAC->SA1 = (params->mac_addr[3] << 8) | params->mac_addr[2];
  LPC_EMAC->SA2 = (params->mac_addr[1] << 8) | params->mac_addr[0];

  if((uintptr_t)dev->mem_area & 7)
  {
    /* Wrong alignment. Abort. */
    return false;
  }

  /* Initialize descriptor pointers. */
  offset = 0;
  dev->rx_descriptors = (lpc17xx_eth_descr_t*) (dev->mem_area + offset);
  offset += dev->rx_descriptors_num * sizeof(lpc17xx_eth_descr_t);
  dev->rx_status = (lpc17xx_rx_status_t*) (dev->mem_area + offset);
  offset += dev->rx_descriptors_num * sizeof(lpc17xx_rx_status_t);
  dev->tx_descriptors = (lpc17xx_eth_descr_t*) (dev->mem_area + offset);
  offset += dev->tx_descriptors_num * sizeof(lpc17xx_eth_descr_t);
  dev->tx_status = (uint32_t*) (dev->mem_area + offset);
  offset += dev->tx_descriptors_num * sizeof(uint32_t);

  /* Initialize descriptors. */
  lpc17xx_eth_init_descriptors(dev, dev->mem_area + offset, dev->mem_sz - offset);

  /* Set EMAC Receive Descriptor Registers. */
  LPC_EMAC->RxDescriptor       = (uint32_t) dev->rx_descriptors;
  LPC_EMAC->RxStatus           = (uint32_t) dev->rx_status;
  LPC_EMAC->RxDescriptorNumber = dev->rx_descriptors_num-1;
  LPC_EMAC->RxConsumeIndex     = 0;

  /* Set EMAC Transmit Descriptor Registers. */
  LPC_EMAC->TxDescriptor       = (uint32_t) dev->tx_descriptors;
  LPC_EMAC->TxStatus           = (uint32_t) dev->tx_status;
  LPC_EMAC->TxDescriptorNumber = dev->tx_descriptors_num-1;
  LPC_EMAC->TxProduceIndex     = 0;

  /* Receive Broadcast and Perfect Match Packets */
  LPC_EMAC->RxFilterCtrl = RFC_BCAST_EN | RFC_PERFECT_EN;

  /* Enable EMAC interrupts. */
  LPC_EMAC->IntEnable = INT_RX_DONE | INT_TX_DONE;

  /* Reset all interrupts */
  LPC_EMAC->IntClear  = 0xFFFF;

  /* Enable receive and transmit mode of MAC Ethernet core */
  LPC_EMAC->Command  |= (CR_RX_EN | CR_TX_EN);
  LPC_EMAC->MAC1     |= MAC1_REC_EN;

  fx_dpc_init(&dev->dpc);
  return true;
}

//----------------------------------------------------------------------------------------------------

static void lpc17xx_eth_dpc_handler(void* dpc, void* arg);

void 
lpc17xx_eth_isr_handler(lpc17xx_eth_dev_t* dev)
{
  const uint32_t int_status = LPC_EMAC->IntStatus;

  if(int_status & INT_TX_DONE)
  {
    if(++dev->tx_consumed == dev->tx_descriptors_num) dev->tx_consumed = 0;
  }

  if(int_status & INT_RX_DONE)
  {
    if(LPC_EMAC->RxProduceIndex != LPC_EMAC->RxConsumeIndex)
    {
      ++dev->rx_service_req;
    }
  } 

  if(dev->rx_service_req == 1)
  {
    fx_dpc_request(&dev->dpc, lpc17xx_eth_dpc_handler, dev);
  }      

  LPC_EMAC->IntClear = int_status;
}

//----------------------------------------------------------------------------------------------------

static void 
lpc17xx_eth_dpc_handler(void* dpc, void* arg)
{
  lpc17xx_eth_dev_t* const dev = (lpc17xx_eth_dev_t*) arg;
  fx_lock_intr_state_t state;
  bool req_present;
  
  do
  {
    dev->rx_signal(dev->rx_arg);
    
    fx_spl_raise_to_sync_from_any(&state);

    req_present = --dev->rx_service_req != 0;

    fx_spl_lower_to_any_from_sync(state);    
  }
  while(req_present);
}

//----------------------------------------------------------------------------------------------------

void 
lpc17xx_eth_deinit(lpc17xx_eth_dev_t* dev)
{
  fx_sched_state_t state;
  fx_sched_lock(&state);

  //(void)fx_isr_detach(&dev->isr);
  (void)fx_dpc_cancel(&dev->dpc);

  fx_sched_unlock(state);  
}

//----------------------------------------------------------------------------------------------------

bool 
lpc17xx_eth_set_mac_addr(lpc17xx_eth_dev_t* dev, const uint8_t* mac)
{
  return false;
}

//----------------------------------------------------------------------------------------------------

bool 
lpc17xx_eth_get_mac_addr(lpc17xx_eth_dev_t* dev, uint8_t* mac)
{
  uint16_t sa0 = LPC_EMAC->SA0;
  uint16_t sa1 = LPC_EMAC->SA1;
  uint16_t sa2 = LPC_EMAC->SA2;

  mac[0] = (uint8_t) sa2;
  mac[1] = sa2 >> 8;
  mac[2] = (uint8_t) sa1;
  mac[3] = sa1 >> 8;
  mac[4] = (uint8_t) sa0;
  mac[5] = sa0 >> 8;

  return true;
}

//----------------------------------------------------------------------------------------------------

bool 
lpc17xx_eth_is_ready_for_tx(lpc17xx_eth_dev_t* dev)
{
  unsigned produce = LPC_EMAC->TxProduceIndex;
  unsigned consume = dev->tx_consumed;
  return ((produce + 1) % dev->tx_descriptors_num) != consume;
}

//----------------------------------------------------------------------------------------------------

bool 
lpc17xx_eth_is_ready_for_rx(lpc17xx_eth_dev_t* dev)
{
  unsigned produce = LPC_EMAC->RxProduceIndex;
  unsigned consume = LPC_EMAC->RxConsumeIndex;
  return (produce != consume);
}

//----------------------------------------------------------------------------------------------------

bool 
lpc17xx_eth_acquire_tx(lpc17xx_eth_dev_t* dev, void** frame, size_t* sz)
{
  const uint32_t idx = LPC_EMAC->TxProduceIndex;
  *frame = (void*) dev->tx_descriptors[idx].buffer;
  *sz = dev->max_frame_len; 
  return true; 
}

//----------------------------------------------------------------------------------------------------

bool 
lpc17xx_eth_send(lpc17xx_eth_dev_t* dev, void* frame, size_t sz)
{
  uint32_t idx = LPC_EMAC->TxProduceIndex;

  dev->tx_descriptors[idx].buffer = (uint32_t) frame;
  dev->tx_descriptors[idx].control = (sz-1) | TCTRL_LAST | TCTRL_INT;

  if((++idx) == dev->tx_descriptors_num) idx = 0;
  LPC_EMAC->TxProduceIndex = idx;  
  return true;
}

//----------------------------------------------------------------------------------------------------

bool 
lpc17xx_eth_receive(lpc17xx_eth_dev_t* dev, void** frame, size_t* sz)
{
  const uint32_t idx = LPC_EMAC->RxConsumeIndex;
  *frame = (void*) dev->rx_descriptors[idx].buffer;
  *sz = (dev->rx_status[idx].status & RCTRL_SIZE) - 3;
  return true;
}

//----------------------------------------------------------------------------------------------------

bool 
lpc17xx_eth_release_rx(lpc17xx_eth_dev_t* dev, void* frame, size_t sz)
{
  uint32_t idx = LPC_EMAC->RxConsumeIndex;
  if(++idx == dev->rx_descriptors_num) idx = 0;
  LPC_EMAC->RxConsumeIndex = idx;   
  return true;
}
