/**
**************************************************************************************************
* @file   ethernetif.c
* @brief  Interface of Ethernet driver for STM32 to LwIP
**************************************************************************************************
* Copyright(C) 2020 Eremex
* All rights reserved.
* ************************************************************************************************/

#include "stm32f4xx_hal.h"

#include "lwip/opt.h"
#include "lwip/timers.h"
#include "netif/etharp.h"

#include "ethernetif.h"

#include "FXRTOS.h"

#include <string.h>


// The time to block waiting for input.
#define TIME_WAITING_FOR_INPUT ( 100 )
// Stack size of the interface thread
#define INTERFACE_THREAD_STACK_SIZE ( 1024 )
uint32_t thread_stack[INTERFACE_THREAD_STACK_SIZE];

// Network interface name
#define IFNAME0 'f'
#define IFNAME1 'x'


ETH_DMADescTypeDef  DMARxDscrTab[ETH_RXBUFNB];	// Ethernet Rx MA Descriptor
ETH_DMADescTypeDef  DMATxDscrTab[ETH_TXBUFNB];	// Ethernet Tx DMA Descriptor
uint8_t Rx_Buff[ETH_RXBUFNB][ETH_RX_BUF_SIZE]; 	// Ethernet Receive Buffer
uint8_t Tx_Buff[ETH_TXBUFNB][ETH_TX_BUF_SIZE]; 	// Ethernet Transmit Buffer


fx_dpc_t      dpc_LwIP;
fx_sem_t      sem_LwIP;
fx_thread_t   thread_LwIP;
//fx_isr_t      isr_LwIP;


// Global Ethernet handle
ETH_HandleTypeDef heth;

const uint8_t DefMACAddr[] = {
	0x02,
	0x00,
	0x00,
	0x00,
	0x00,
	0x07
};

//------------------------------------------------------------------------------
//
// Peripheral clock enable
//
void HAL_ETH_MspInit(ETH_HandleTypeDef* heth)
{
  GPIO_InitTypeDef GPIO_InitStruct;

  if(heth->Instance==ETH)
  {


    __ETH_CLK_DISABLE();
    HAL_Delay(10);

    /**ETH GPIO Configuration    
    PC1     ------> ETH_MDC
    PA1     ------> ETH_REF_CLK
    PA2     ------> ETH_MDIO
    PA7     ------> ETH_CRS_DV
    PC4     ------> ETH_RXD0
    PC5     ------> ETH_RXD1
    PB11     ------> ETH_TX_EN
    PB12     ------> ETH_TXD0
    PB13     ------> ETH_TXD1 
    */
    GPIO_InitStruct.Pin = GPIO_PIN_1|GPIO_PIN_4|GPIO_PIN_5;
    GPIO_InitStruct.Mode = GPIO_MODE_AF_PP;
    GPIO_InitStruct.Pull = GPIO_NOPULL;
    GPIO_InitStruct.Speed = GPIO_SPEED_LOW;
    GPIO_InitStruct.Alternate = GPIO_AF11_ETH;
    HAL_GPIO_Init(GPIOC, &GPIO_InitStruct);

    GPIO_InitStruct.Pin = GPIO_PIN_1|GPIO_PIN_2|GPIO_PIN_7;
    GPIO_InitStruct.Mode = GPIO_MODE_AF_PP;
    GPIO_InitStruct.Pull = GPIO_NOPULL;
    GPIO_InitStruct.Speed = GPIO_SPEED_LOW;
    GPIO_InitStruct.Alternate = GPIO_AF11_ETH;
    HAL_GPIO_Init(GPIOA, &GPIO_InitStruct);

    GPIO_InitStruct.Pin = GPIO_PIN_11|GPIO_PIN_12|GPIO_PIN_13;
    GPIO_InitStruct.Mode = GPIO_MODE_AF_PP;
    GPIO_InitStruct.Pull = GPIO_NOPULL;
    GPIO_InitStruct.Speed = GPIO_SPEED_LOW;
    GPIO_InitStruct.Alternate = GPIO_AF11_ETH;
    HAL_GPIO_Init(GPIOB, &GPIO_InitStruct);


    //Configure GPIO pin : PE2
    GPIO_InitStruct.Pin = GPIO_PIN_2;
    GPIO_InitStruct.Mode = GPIO_MODE_OUTPUT_PP;
    GPIO_InitStruct.Pull = GPIO_NOPULL;
    GPIO_InitStruct.Speed = GPIO_SPEED_LOW;
    HAL_GPIO_Init(GPIOE, &GPIO_InitStruct);

    HAL_GPIO_WritePin(GPIOE, GPIO_PIN_2, GPIO_PIN_RESET);
    HAL_Delay(10);

    __ETH_CLK_ENABLE();
    HAL_Delay(10);

    HAL_GPIO_WritePin(GPIOE, GPIO_PIN_2, GPIO_PIN_SET);	
    HAL_Delay(10);
  }
}

//------------------------------------------------------------------------------
//
// Peripheral clock disable
//
void HAL_ETH_MspDeInit(ETH_HandleTypeDef* heth)
{
  if(heth->Instance==ETH)
  {
    // Peripheral clock disable
    __ETH_CLK_DISABLE();
  

    HAL_GPIO_DeInit(GPIOC, GPIO_PIN_1|GPIO_PIN_4|GPIO_PIN_5);
    HAL_GPIO_DeInit(GPIOA, GPIO_PIN_1|GPIO_PIN_2|GPIO_PIN_7);
    HAL_GPIO_DeInit(GPIOB, GPIO_PIN_11|GPIO_PIN_12|GPIO_PIN_13);
    HAL_GPIO_DeInit(GPIOE, GPIO_PIN_2);

    // Peripheral interrupt Deinit
    //fx_isr_detach(&isr_LwIP);

  }
}

//------------------------------------------------------------------------------
//
// @brief  Ethernet Rx Transfer completed callback
//

void dpc_func_LwIP(void* dpc, void* arg)
{
	fx_sem_post(&sem_LwIP);
}

void HAL_ETH_RxCpltCallback(ETH_HandleTypeDef *heth)
{
	fx_dpc_request(&dpc_LwIP, dpc_func_LwIP, (void*)0x0);
}

bool isr_func_LwIP(void* arg)
{
	HAL_ETH_IRQHandler(&heth);
  return true;
}

//------------------------------------------------------------------------------
//                       LL Driver Interface ( LwIP stack --> ETH)
//
// In this function, the hardware should be initialized.
// Called from ethernetif_init().
//
// @param netif the already initialized lwip network interface structure for this ethernetif
//
static void low_level_init(struct netif *netif)
{ 
  uint32_t regvalue = 0;
  HAL_StatusTypeDef hal_eth_init_status;

  // Init ETH

  heth.Instance = ETH;
  heth.Init.AutoNegotiation = ETH_AUTONEGOTIATION_ENABLE;
  heth.Init.Speed = ETH_SPEED_100M;
  heth.Init.DuplexMode = ETH_MODE_FULLDUPLEX;
  heth.Init.PhyAddress = LAN8720_PHY_ADDRESS;
  heth.Init.MACAddr = DefMACAddr;
  heth.Init.RxMode = ETH_RXINTERRUPT_MODE;
  heth.Init.ChecksumMode = ETH_CHECKSUM_BY_SOFTWARE;
  heth.Init.MediaInterface = ETH_MEDIA_INTERFACE_RMII;
  hal_eth_init_status = HAL_ETH_Init(&heth);

  if (hal_eth_init_status == HAL_OK)
  {
    // Set netif link flag
    netif->flags |= NETIF_FLAG_LINK_UP;
  }

  // Initialize Tx Descriptors list: Chain Mode
  HAL_ETH_DMATxDescListInit(&heth, DMATxDscrTab, &Tx_Buff[0][0], ETH_TXBUFNB);
     
  // Initialize Rx Descriptors list: Chain Mode
  HAL_ETH_DMARxDescListInit(&heth, DMARxDscrTab, &Rx_Buff[0][0], ETH_RXBUFNB);

#if LWIP_ARP || LWIP_ETHERNET 
  // set MAC hardware address length
  netif->hwaddr_len = ETHARP_HWADDR_LEN;
  
  // set MAC hardware address
  netif->hwaddr[0] =  heth.Init.MACAddr[0];
  netif->hwaddr[1] =  heth.Init.MACAddr[1];
  netif->hwaddr[2] =  heth.Init.MACAddr[2];
  netif->hwaddr[3] =  heth.Init.MACAddr[3];
  netif->hwaddr[4] =  heth.Init.MACAddr[4];
  netif->hwaddr[5] =  heth.Init.MACAddr[5];
  
  // maximum transfer unit
  netif->mtu = 1500;
  
  // Accept broadcast address and ARP traffic
  // don't set NETIF_FLAG_ETHARP if this device is not an ethernet one
  #if LWIP_ARP
    netif->flags |= NETIF_FLAG_BROADCAST | NETIF_FLAG_ETHARP;
  #else 
    netif->flags |= NETIF_FLAG_BROADCAST;
  #endif // LWIP_ARP
  

	fx_dpc_init(&dpc_LwIP);
	fx_sem_init(&sem_LwIP, 0, 100000, FX_SYNC_POLICY_DEFAULT);
	fx_thread_init(&thread_LwIP, ethernetif_input, (void*)netif, 1, (void*)thread_stack, sizeof(thread_stack), false);
	//fx_isr_init(&isr_LwIP, ETH_IRQn, isr_func_LwIP, NULL);
	//fx_isr_attach(&isr_LwIP);


  // Enable MAC and DMA transmission and reception
  HAL_ETH_Start(&heth);
  
  //  Configure PHY to generate an interrupt when Eth Link state changes
  // Read Register Configuration
  HAL_ETH_ReadPHYRegister(&heth, PHY_MICR, &regvalue);
  
  regvalue |= (PHY_MICR_INT_EN | PHY_MICR_INT_OE);

  // Enable Interrupts
  HAL_ETH_WritePHYRegister(&heth, PHY_MICR, regvalue );
  
  // Read Register Configuration
  HAL_ETH_ReadPHYRegister(&heth, PHY_MISR, &regvalue);
  
  regvalue |= PHY_MISR_LINK_INT_EN;
    
  // Enable Interrupt on change of link status
  HAL_ETH_WritePHYRegister(&heth, PHY_MISR, regvalue);   
#endif /* LWIP_ARP || LWIP_ETHERNET */
}

//------------------------------------------------------------------------------
//
// This function should do the actual transmission of the packet. The packet is
// contained in the pbuf that is passed to the function. This pbuf
// might be chained.
//
// @param netif the lwip network interface structure for this ethernetif
// @param p the MAC packet to send (e.g. IP packet including MAC addresses and type)
// @return ERR_OK if the packet could be sent
//         an err_t value if the packet couldn't be sent
//
// @note Returning ERR_MEM here if a DMA queue of your MAC is full can lead to
//       strange results. You might consider waiting for space in the DMA queue
//       to become availale since the stack doesn't retry to send a packet
//       dropped because of memory failure (except for the TCP timers).
//
static err_t low_level_output(struct netif *netif, struct pbuf *p)
{
  err_t errval;
  struct pbuf *q;
  uint8_t *buffer = (uint8_t *)(heth.TxDesc->Buffer1Addr);
  __IO ETH_DMADescTypeDef *DmaTxDesc;
  uint32_t framelength = 0;
  uint32_t bufferoffset = 0;
  uint32_t byteslefttocopy = 0;
  uint32_t payloadoffset = 0;
  DmaTxDesc = heth.TxDesc;
  bufferoffset = 0;
  
  // copy frame from pbufs to driver buffers
  for(q = p; q != NULL; q = q->next)
  {
      // Is this buffer available? If not, goto error
      if((DmaTxDesc->Status & ETH_DMATXDESC_OWN) != (uint32_t)RESET)
      {
        errval = ERR_USE;
        goto error;
      }
    
      // Get bytes in current lwIP buffer
      byteslefttocopy = q->len;
      payloadoffset = 0;
    
      // Check if the length of data to copy is bigger than Tx buffer size
      while( (byteslefttocopy + bufferoffset) > ETH_TX_BUF_SIZE )
      {
        // Copy data to Tx buffer
        memcpy( (uint8_t*)((uint8_t*)buffer + bufferoffset), (uint8_t*)((uint8_t*)q->payload + payloadoffset), (ETH_TX_BUF_SIZE - bufferoffset) );
      
        // Point to next descriptor
        DmaTxDesc = (ETH_DMADescTypeDef *)(DmaTxDesc->Buffer2NextDescAddr);
      
        // Check if the buffer is available
        if((DmaTxDesc->Status & ETH_DMATXDESC_OWN) != (uint32_t)RESET)
        {
          errval = ERR_USE;
          goto error;
        }
      
        buffer = (uint8_t *)(DmaTxDesc->Buffer1Addr);
      
        byteslefttocopy = byteslefttocopy - (ETH_TX_BUF_SIZE - bufferoffset);
        payloadoffset = payloadoffset + (ETH_TX_BUF_SIZE - bufferoffset);
        framelength = framelength + (ETH_TX_BUF_SIZE - bufferoffset);
        bufferoffset = 0;
      }
    
      // Copy the remaining bytes
      memcpy( (uint8_t*)((uint8_t*)buffer + bufferoffset), (uint8_t*)((uint8_t*)q->payload + payloadoffset), byteslefttocopy );
      bufferoffset = bufferoffset + byteslefttocopy;
      framelength = framelength + byteslefttocopy;
  }
  
  // Prepare transmit descriptors to give to DMA
  HAL_ETH_TransmitFrame(&heth, framelength);
  
  errval = ERR_OK;
  
error:
  
  // When Transmit Underflow flag is set, clear it and issue a Transmit Poll Demand to resume transmission
  if ((heth.Instance->DMASR & ETH_DMASR_TUS) != (uint32_t)RESET)
  {
    // Clear TUS ETHERNET DMA flag
    heth.Instance->DMASR = ETH_DMASR_TUS;

    // Resume DMA transmission
    heth.Instance->DMATPDR = 0;
  }
  return errval;
}

//------------------------------------------------------------------------------
//
// Should allocate a pbuf and transfer the bytes of the incoming
// packet from the interface into the pbuf.
//
// @param netif the lwip network interface structure for this ethernetif
// @return a pbuf filled with the received packet (including MAC header)
//         NULL on memory error
//
static struct pbuf * low_level_input(struct netif *netif)
{
  struct pbuf *p = NULL;
  struct pbuf *q;
  uint16_t len = 0;
  uint8_t *buffer;
  __IO ETH_DMADescTypeDef *dmarxdesc;
  uint32_t bufferoffset = 0;
  uint32_t payloadoffset = 0;
  uint32_t byteslefttocopy = 0;
  uint32_t i=0;
  

  // get received frame
  if (HAL_ETH_GetReceivedFrame_IT(&heth) != HAL_OK)
    return NULL;
  
  // Obtain the size of the packet and put it into the "len" variable.
  len = heth.RxFrameInfos.length;
  buffer = (uint8_t *)heth.RxFrameInfos.buffer;
  
  if (len > 0)
  {
    // We allocate a pbuf chain of pbufs from the Lwip buffer pool
    p = pbuf_alloc(PBUF_RAW, len, PBUF_POOL);
  }
  
  if (p != NULL)
  {
    dmarxdesc = heth.RxFrameInfos.FSRxDesc;
    bufferoffset = 0;
    for(q = p; q != NULL; q = q->next)
    {
      byteslefttocopy = q->len;
      payloadoffset = 0;
      
      // Check if the length of bytes to copy in current pbuf is bigger than Rx buffer size
      while( (byteslefttocopy + bufferoffset) > ETH_RX_BUF_SIZE )
      {
        // Copy data to pbuf
        memcpy( (uint8_t*)((uint8_t*)q->payload + payloadoffset), (uint8_t*)((uint8_t*)buffer + bufferoffset), (ETH_RX_BUF_SIZE - bufferoffset));
        
        // Point to next descriptor
        dmarxdesc = (ETH_DMADescTypeDef *)(dmarxdesc->Buffer2NextDescAddr);
        buffer = (uint8_t *)(dmarxdesc->Buffer1Addr);
        
        byteslefttocopy = byteslefttocopy - (ETH_RX_BUF_SIZE - bufferoffset);
        payloadoffset = payloadoffset + (ETH_RX_BUF_SIZE - bufferoffset);
        bufferoffset = 0;
      }
      // Copy remaining data in pbuf
      memcpy( (uint8_t*)((uint8_t*)q->payload + payloadoffset), (uint8_t*)((uint8_t*)buffer + bufferoffset), byteslefttocopy);
      bufferoffset = bufferoffset + byteslefttocopy;
    }
    
    // Release descriptors to DMA
    // Point to first descriptor
    dmarxdesc = heth.RxFrameInfos.FSRxDesc;
    // Set Own bit in Rx descriptors: gives the buffers back to DMA
    for (i=0; i< heth.RxFrameInfos.SegCount; i++)
    {  
      dmarxdesc->Status |= ETH_DMARXDESC_OWN;
      dmarxdesc = (ETH_DMADescTypeDef *)(dmarxdesc->Buffer2NextDescAddr);
    }
    
    // Clear Segment_Count
    heth.RxFrameInfos.SegCount =0;
  }    
  
  // When Rx Buffer unavailable flag is set: clear it and resume reception
  if ((heth.Instance->DMASR & ETH_DMASR_RBUS) != (uint32_t)RESET)  
  {
    // Clear RBUS ETHERNET DMA flag
    heth.Instance->DMASR = ETH_DMASR_RBUS;
    // Resume DMA reception
    heth.Instance->DMARPDR = 0;
  }
  return p;
}

//------------------------------------------------------------------------------
//
// This function should be called when a packet is ready to be read
// from the interface. It uses the function low_level_input() that
// should handle the actual reception of bytes from the network
// interface. Then the type of the received packet is determined and
// the appropriate input function is called.
//
// @param netif the lwip network interface structure for this ethernetif
//
void ethernetif_input( void* argument )
{
  struct pbuf *p;
  struct netif *netif = (struct netif *) argument;
  
  for( ;; )
  {
    if (fx_sem_timedwait(&sem_LwIP, TIME_WAITING_FOR_INPUT) == FX_SEM_OK)
    {
      do
      {   
        p = low_level_input( netif );
        if   (p != NULL)
        {
          if (netif->input( p, netif) != ERR_OK )
          {
            pbuf_free(p);
          }
        }
      } while(p!=NULL);
    }
  }
}


//------------------------------------------------------------------------------
//
// Should be called at the beginning of the program to set up the
// network interface. It calls the function low_level_init() to do the
// actual setup of the hardware.
//
// This function should be passed as a parameter to netif_add().
//
// @param netif the lwip network interface structure for this ethernetif
// @return ERR_OK if the loopif is initialized
//         ERR_MEM if private data couldn't be allocated
//         any other err_t on error
//
err_t ethernetif_init(struct netif *netif)
{
  LWIP_ASSERT("netif != NULL", (netif != NULL));
  
#if LWIP_NETIF_HOSTNAME
  /* Initialize interface hostname */
  netif->hostname = "lwip";
#endif /* LWIP_NETIF_HOSTNAME */

  netif->name[0] = IFNAME0;
  netif->name[1] = IFNAME1;
  // We directly use etharp_output() here to save a function call.
  // You can instead declare your own function an call etharp_output()
  // from it if you have to do some checks before sending (e.g. if link
  // is available...)
#if LWIP_ARP || LWIP_ETHERNET
#if LWIP_ARP
  netif->output = etharp_output;
#else
  // The user should write ist own code in low_level_output_arp_off function
  netif->output = low_level_output_arp_off;
#endif /* LWIP_ARP */ 
#endif  /* LWIP_ARP || LWIP_ETHERNET */
  netif->linkoutput = low_level_output;

  // initialize the hardware
  low_level_init(netif);

  return ERR_OK;
}


//------------------------------------------------------------------------------
uint32_t HAL_GetTick(void)
{
  return fx_timer_get_tick_count();
}
