/**
  ******************************************************************************
  *  @file   stm32_bitband_nand.c
  *  @brief  Bit-banging NAND memory interface for stm32f411re.
  *  Tested on IC:
  *  SAMSUNG K9F1G08U08B 65,536 row (pages), by 2,048x8 columns,
  *  spare 64x8 columns
  *
  ******************************************************************************
  *  Copyright (C) 2021 Eremex
  *  All rights reserved.
  *****************************************************************************/

#include <nand.h>
#include <hamming.h>

#include "FXRTOS.h"
#include "fx_nand_hw.h"

#include "stm32f4xx.h"
#include "stm32f4xx_ll_rcc.h"
#include "stm32f4xx_ll_bus.h"
#include "stm32f4xx_ll_system.h"
#include "stm32f4xx_ll_exti.h"
#include "stm32f4xx_ll_cortex.h"
#include "stm32f4xx_ll_tim.h"
#include "stm32f4xx_ll_gpio.h"

#define NAND_D0_Pin LL_GPIO_PIN_0
#define NAND_D0_GPIO_Port GPIOA
#define NAND_D1_Pin LL_GPIO_PIN_1
#define NAND_D1_GPIO_Port GPIOA
#define NAND_D2_Pin LL_GPIO_PIN_2
#define NAND_D2_GPIO_Port GPIOA
#define NAND_D3_Pin LL_GPIO_PIN_3
#define NAND_D3_GPIO_Port GPIOA
#define NAND_D4_Pin LL_GPIO_PIN_4
#define NAND_D4_GPIO_Port GPIOA
#define NAND_D5_Pin LL_GPIO_PIN_5
#define NAND_D5_GPIO_Port GPIOA
#define NAND_D6_Pin LL_GPIO_PIN_6
#define NAND_D6_GPIO_Port GPIOA
#define NAND_D7_Pin LL_GPIO_PIN_7
#define NAND_D7_GPIO_Port GPIOA
#define NAND_nWP_Pin LL_GPIO_PIN_0
#define NAND_nWP_GPIO_Port GPIOB
#define NAND_nRB_Pin LL_GPIO_PIN_1
#define NAND_nRB_GPIO_Port GPIOB
#define NAND_ALE_Pin LL_GPIO_PIN_8
#define NAND_ALE_GPIO_Port GPIOA
#define NAND_CLE_Pin LL_GPIO_PIN_9
#define NAND_CLE_GPIO_Port GPIOA
#define NAND_nWE_Pin LL_GPIO_PIN_10
#define NAND_nWE_GPIO_Port GPIOA
#define NAND_nRE_Pin LL_GPIO_PIN_11
#define NAND_nRE_GPIO_Port GPIOA
#define NAND_nCE_Pin LL_GPIO_PIN_12
#define NAND_nCE_GPIO_Port GPIOA

#define PAGE_SIZE       2048
#define SPARE_SIZE      64

#define PPB             64
#define TOTAL_PAGES     65536
#define TOTAL_BLOCKS    1024

#define COL_BIT_WIDTH   11
#define PPB_BIT_WIDTH   6
#define BUS_WIDTH       8

static unsigned chipselect = Chip0_Select;

fx_sem_t nand_ready_sem;
static int read_trace_val = 0;
//
// Deferred interrupt processing function.
//
void nand_irq_dpc_handler(fx_dpc_t *dpc, void *arg)
{
    //
    // Post semaphore and wake the waiting thread up.
    //
    fx_sem_post(&nand_ready_sem);
}

uint8_t nand_status(NAND_Chip chipselect)
{
    nand_chip_enable(chipselect);
    nand_set_dir_output();
    nand_write_cmd(NAND_CMD_READ_STATUS);
    nand_set_dir_input();
    uint8_t state = nand_read_byte();
    nand_chip_enable(~chipselect);

    return state;
}

//
// Program the given page. The data pointer is a pointer to an entire
// page ((1 << log2_page_size) bytes). The operation status should be
// checked. If the operation fails, return -1 and set err to
// E_BAD_BLOCK.
//
// Pages will be programmed sequentially within a block, and will not be
// reprogrammed.
//
int dhara_nand_prog(const struct dhara_nand *n, dhara_page_t page_addr,
        const uint8_t *data, dhara_error_t *err)
{
    //
    // Compute ECC
    //
    uint8_t ecc[SPARE_SIZE];
    hamming_generate(data, PAGE_SIZE, ecc);

    nand_chip_enable(chipselect);
    nand_set_dir_output();

    nand_write_cmd(NAND_CMD_PROGRAM_PAGE_CYCLE1);

    //
    // Write starting column address 0
    //
    int col_addr_bits = n->log2_page_size;
    while (col_addr_bits > 0)
    {
        nand_write_addr(0);
        col_addr_bits -= BUS_WIDTH;
    };

    //
    // Write row address: count bit length of row address and write by 8-bit
    //
    int row_width = __builtin_ctz(n->num_blocks * (1 << (n->log2_ppb)));
    while (row_width > 0)
    {
        nand_write_addr(page_addr & 0xFF);
        page_addr >>= BUS_WIDTH;
        row_width -= BUS_WIDTH;
    };

    //
    // Write data
    //
    for (const uint8_t *p = data; p < data + PAGE_SIZE; p++)
    {
        nand_write_byte(*p);
    }

    //
    // Write ECC
    //
    for (const uint8_t *p = ecc; p < ecc + SPARE_SIZE; p++)
    {
       nand_write_byte(*p);
    }

    //
    // 10h PROGRAM command to confirm and start programming operation
    //
    nand_write_cmd(NAND_CMD_PROGRAM_PAGE_CYCLE2);

    while (!nand_get_ready(chipselect))
    {
        ;
    }
    nand_chip_enable(~chipselect);

    int ret = 0;
    if ((nand_status(chipselect) & NAND_ERASE_PROGRAM_FAIL) != 0)
    {
        ret = -1;
        *err = DHARA_E_BAD_BLOCK;
    }

    return ret;
}

//
// Read a portion of a page. ECC must be handled by the NAND
// implementation. Returns 0 on sucess or -1 if an error occurs. If an
// uncorrectable ECC error occurs, return -1 and set err to E_ECC.
//
int dhara_nand_read(const struct dhara_nand *n, dhara_page_t page_addr,
        size_t offset, size_t length, uint8_t *data, dhara_error_t *err)
{
    read_trace_val = 1;

    nand_set_dir_output();

    nand_chip_enable(chipselect);
    nand_write_cmd(NAND_CMD_READ_PAGE_CYCLE1);

    //
    // Write starting column address
    //
    int col_addr_bits = n->log2_page_size;
    while (col_addr_bits > 0)
    {
        nand_write_addr(offset & 0xFF);
        offset >>= BUS_WIDTH;
        col_addr_bits -= BUS_WIDTH;
    };

    //
    // Write row address: count bit length of row address and write by 8-bit
    //
    int row_width = __builtin_ctz(n->num_blocks * (1 << (n->log2_ppb)));
    while (row_width > 0)
    {
        nand_write_addr(page_addr & 0xFF);
        page_addr >>= BUS_WIDTH;
        row_width -= BUS_WIDTH;
    };

    nand_write_cmd(NAND_CMD_READ_PAGE_CYCLE2);

    nand_set_dir_input();

    while (!nand_get_ready(chipselect))
    {
        ;
    }

    nand_set_dir_input();
    for (uint8_t *p = data; p < data + length; p++)
    {
        *p = nand_read_byte();
    }

    int ret = 0;
    if (length == PAGE_SIZE)
    {
        uint8_t ecc[SPARE_SIZE];
        //
        // Read ECC
        //
        for (uint8_t *p = ecc; p < ecc + SPARE_SIZE; p++)
        {
            *p = nand_read_byte();
        }
        //
        // Check ECC, and try to recover data in buffer. If repair failed set err to E_ECC.
        // Return -1 if at least recoverable error found.
        //
        hamming_ecc_t syndr = hamming_syndrome(data, PAGE_SIZE, ecc);
        if (syndr != 0)
        {
            ret = -1;
            if (0 != hamming_repair(data, PAGE_SIZE, syndr))
            {
                *err = DHARA_E_ECC;
            }
        }
    }

    nand_chip_enable(~chipselect);
    read_trace_val = 0;

    return ret;
}

//
// Erase the given block. This function should return 0 on success or -1
// on failure.
//
// The status reported by the chip should be checked. If an erase
// operation fails, return -1 and set err to E_BAD_BLOCK.
//
int dhara_nand_erase(const struct dhara_nand *n, dhara_block_t b,
        dhara_error_t *err)
{
    nand_chip_enable(chipselect);

    nand_set_dir_output();
    nand_write_cmd(NAND_CMD_ERASE_BLOCK_CYCLE1);

    //
    // Write row address (eraseblock address)
    //
    int row_width = __builtin_ctz(n->num_blocks * (1 << (n->log2_ppb)));
    uint32_t block_addr = b << (n->log2_ppb);
    while (row_width > 0)
    {
        nand_write_addr(block_addr & 0xFF);
        block_addr >>= BUS_WIDTH;
        row_width -= BUS_WIDTH;
    };

    nand_write_cmd(NAND_CMD_ERASE_BLOCK_CYCLE2);
    while (!nand_get_ready(chipselect))
    {
        ;
    }
    nand_chip_enable(~chipselect);

    int ret = 0;
    if ((nand_status(chipselect) & NAND_ERASE_PROGRAM_FAIL) != 0)
    {
        ret = -1;
        *err = DHARA_E_BAD_BLOCK;
    }

    return ret;
}

//
// Check that the given page is erased
//
int dhara_nand_is_free(const struct dhara_nand *n, dhara_page_t p)
{
    int is_free = true;
    uint8_t buf[PAGE_SIZE + SPARE_SIZE];
    dhara_nand_read(n, p, 0, PAGE_SIZE + SPARE_SIZE, buf, NULL);

    for (size_t k = 0; k < PAGE_SIZE + SPARE_SIZE; k++)
    {
        if (buf[k] != 0xFF)
        {
            is_free = false;
            break;
        }
    }
    return is_free;
}

//
// Read a page from one location and reprogram it in another location.
// This might be done using the chip's internal buffers, but it must use
// ECC.
//
int dhara_nand_copy(const struct dhara_nand *n, dhara_page_t src,
        dhara_page_t dst, dhara_error_t *err)
{
    uint8_t buf[PAGE_SIZE];
    //
    // Read page with ECC
    //
    int ret = dhara_nand_read(n, src, 0, PAGE_SIZE, buf, err);
    if (*err != DHARA_E_ECC)
    {
        //
        // Program only unbroken or recovered data
        //
        ret = dhara_nand_prog(n, dst, buf, NULL);
    }

    return ret;
}

//
// Is the given block bad?
//
int dhara_nand_is_bad(const struct dhara_nand *n, dhara_block_t b)
{
    int is_bad = false;

    dhara_page_t page;
    dhara_page_t first_page_in_block = b << PPB_BIT_WIDTH;
    for (page = first_page_in_block; page <= first_page_in_block + 1; page++)
    {
        uint8_t bad_marker;
        dhara_nand_read(n, page, PAGE_SIZE, 1, &bad_marker, NULL);
        if (bad_marker != 0xFF)
        {
            is_bad = true;
            break;
        }
    }

    return is_bad;
}

//
// Mark bad the given block (or attempt to). No return value is
// required, because there's nothing that can be done in response.
//
void dhara_nand_mark_bad(const struct dhara_nand *n, dhara_block_t b)
{
    uint8_t bad_marker = 0xBD;
    dhara_page_t page0_in_b = b << PPB_BIT_WIDTH;

    //
    // We should erase dirty block before program
    //
    dhara_nand_erase(n, b, NULL);

    nand_chip_enable(chipselect);
    nand_set_dir_output();

    nand_write_cmd(NAND_CMD_PROGRAM_PAGE_CYCLE1);

    //
    // Write first byte in the redundant area to mark block as bad
    //
    int col_addr_bits = n->log2_page_size;
    while (col_addr_bits > 0)
    {
        nand_write_addr(0);
        col_addr_bits -= BUS_WIDTH;
    };

    //
    // Write row address: count bit length of row address and write by 8-bit
    //
    int row_width = __builtin_ctz(n->num_blocks * (1 << (n->log2_ppb)));
    while (row_width > 0)
    {
        nand_write_addr(page0_in_b & 0xFF);
        page0_in_b >>= BUS_WIDTH;
        row_width -= BUS_WIDTH;
    };

    //
    // Write data 0xBD
    //
    for (size_t i = 0; i < PAGE_SIZE + SPARE_SIZE; i++)
    {
        nand_write_byte(bad_marker);
    }

    //
    // 10h PROGRAM command to confirm and start programming operation
    //
    nand_write_cmd(NAND_CMD_PROGRAM_PAGE_CYCLE2);

    while (!nand_get_ready(chipselect))
    {
        ;
    }
    nand_chip_enable(~chipselect);
}

void nand_bus_configure(void)
{
    LL_GPIO_InitTypeDef GPIO_InitStruct = { 0 };

    LL_AHB1_GRP1_EnableClock(LL_AHB1_GRP1_PERIPH_GPIOC);
    LL_AHB1_GRP1_EnableClock(LL_AHB1_GRP1_PERIPH_GPIOH);
    LL_AHB1_GRP1_EnableClock(LL_AHB1_GRP1_PERIPH_GPIOA);
    LL_AHB1_GRP1_EnableClock(LL_AHB1_GRP1_PERIPH_GPIOB);

    LL_GPIO_SetOutputPin(NAND_nWP_GPIO_Port, NAND_nWP_Pin);

    LL_GPIO_SetOutputPin(GPIOA,
            NAND_ALE_Pin | NAND_CLE_Pin | NAND_nWE_Pin | NAND_nRE_Pin
                    | NAND_nCE_Pin);

    GPIO_InitStruct.Pin = NAND_D0_Pin | NAND_D1_Pin | NAND_D2_Pin | NAND_D3_Pin
            | NAND_D4_Pin | NAND_D5_Pin | NAND_D6_Pin | NAND_D7_Pin;
    GPIO_InitStruct.Mode = LL_GPIO_MODE_INPUT;
    GPIO_InitStruct.Pull = LL_GPIO_PULL_UP;
    LL_GPIO_Init(GPIOA, &GPIO_InitStruct);

    GPIO_InitStruct.Pin = NAND_nWP_Pin;
    GPIO_InitStruct.Mode = LL_GPIO_MODE_OUTPUT;
    GPIO_InitStruct.Speed = LL_GPIO_SPEED_FREQ_VERY_HIGH;
    GPIO_InitStruct.OutputType = LL_GPIO_OUTPUT_PUSHPULL;
    GPIO_InitStruct.Pull = LL_GPIO_PULL_NO;
    LL_GPIO_Init(NAND_nWP_GPIO_Port, &GPIO_InitStruct);

    GPIO_InitStruct.Pin = NAND_nRB_Pin;
    GPIO_InitStruct.Mode = LL_GPIO_MODE_INPUT;
    GPIO_InitStruct.Pull = LL_GPIO_PULL_UP;
    LL_GPIO_Init(NAND_nRB_GPIO_Port, &GPIO_InitStruct);

    GPIO_InitStruct.Pin = NAND_ALE_Pin | NAND_CLE_Pin | NAND_nWE_Pin
            | NAND_nRE_Pin | NAND_nCE_Pin;
    GPIO_InitStruct.Mode = LL_GPIO_MODE_OUTPUT;
    GPIO_InitStruct.Speed = LL_GPIO_SPEED_FREQ_VERY_HIGH;
    GPIO_InitStruct.OutputType = LL_GPIO_OUTPUT_PUSHPULL;
    GPIO_InitStruct.Pull = LL_GPIO_PULL_NO;
    LL_GPIO_Init(GPIOA, &GPIO_InitStruct);
}

bool nand_get_ready(NAND_Chip cs)
{
    bool ready = false;
    switch (cs)
    {
    case Chip0_Select:
        NVIC_EnableIRQ(EXTI15_10_IRQn);
        fx_sem_timedwait(&nand_ready_sem, FX_THREAD_INFINITE_TIMEOUT);
        ready = (0 != LL_GPIO_IsInputPinSet(NAND_nRB_GPIO_Port, NAND_nRB_Pin));
        break;
    default:
        break;
    }

    return ready;
}

void nand_chip_enable(NAND_Chip cs)
{
    switch (cs)
    {
    case Chip0_Select:
        LL_GPIO_ResetOutputPin(NAND_nCE_GPIO_Port, NAND_nCE_Pin);
        break;
    case Chip0_UnSelect:
        LL_GPIO_SetOutputPin(NAND_nCE_GPIO_Port, NAND_nCE_Pin);
        break;
    default:
        break;
    }
}

//
// Set bus [7..0] to input
// Setup pull-up on pins
//
void nand_set_dir_input(void)
{
    MODIFY_REG(NAND_D0_GPIO_Port->MODER, 0xFFFF, 0x0000);
    MODIFY_REG(NAND_D0_GPIO_Port->PUPDR, 0xFFFF, 0x5555);
}

//
// Set bus [7..0] to output
//
void nand_set_dir_output(void)
{
    MODIFY_REG(NAND_D0_GPIO_Port->MODER, 0xFFFF, 0x5555);
}

void nand_write_cmd(uint8_t cmd)
{
    //
    // CLE _____-------------___
    // WE --------_____---------
    // IO ----------######------
    // t(DS)        ^  ^        = 12 ns min
    // t(DH)           ^  ^     = 5 ns min
    // t(CLS)   ^      ^        = 12 ns min
    // t(CLH)          ^     ^  = 5 ns min
    //
    // CLE=1, ALE=0, nWE=0, nRE=1
    //
    LL_GPIO_WriteOutputPort(NAND_D0_GPIO_Port,
            NAND_CLE_Pin | NAND_nRE_Pin | cmd);
    //
    // Delay
    //
    __NOP();
    __NOP();
    __NOP();
    __NOP();
    __NOP();
    __NOP();
    __NOP();
    __NOP();
    __NOP();
    __NOP();
    //
    // CLE=1, ALE=0, nWE=1, nRE=1
    //
    LL_GPIO_WriteOutputPort(NAND_D0_GPIO_Port,
            NAND_CLE_Pin | NAND_nWE_Pin | NAND_nRE_Pin | cmd);
    __NOP();
    __NOP();
    //
    // CLE=0, ALE=0, nWE=1, nRE=1
    //
    LL_GPIO_WriteOutputPort(NAND_D0_GPIO_Port,
            NAND_nWE_Pin | NAND_nRE_Pin | cmd);
}

void nand_write_addr(uint8_t addr)
{
    //
    // CLE=0, ALE=1, nWE=0, nRE=1
    //
    LL_GPIO_WriteOutputPort(NAND_D0_GPIO_Port,
            NAND_ALE_Pin | NAND_nRE_Pin | addr);
    __NOP();
    __NOP();
    __NOP();
    __NOP();
    __NOP();
    __NOP();
    __NOP();
    __NOP();
    __NOP();
    __NOP();
    //
    // CLE=0, ALE=1, nWE=1, nRE=1
    //
    LL_GPIO_WriteOutputPort(NAND_D0_GPIO_Port,
            NAND_ALE_Pin | NAND_nWE_Pin | NAND_nRE_Pin | addr);
    __NOP();
    __NOP();
    //
    // CLE=0, ALE=0, nWE=1, nRE=1
    //
    LL_GPIO_WriteOutputPort(NAND_D0_GPIO_Port,
            NAND_nWE_Pin | NAND_nRE_Pin | addr);
}

uint8_t nand_read_byte(void)
{
    //
    // CLE=0, ALE=0, nWE=1, nRE=0
    //
    LL_GPIO_WriteOutputPort(NAND_D0_GPIO_Port, NAND_nWE_Pin | 0xFF);
    __NOP();
    __NOP();
    __NOP();
    __NOP();
    __NOP();
    __NOP();
    __NOP();
    __NOP();
    __NOP();
    __NOP();
    uint32_t data = LL_GPIO_ReadInputPort(NAND_D0_GPIO_Port);
    //
    // CLE=0, ALE=0, nWE=1, nRE=1
    //
    LL_GPIO_WriteOutputPort(NAND_D0_GPIO_Port,
            NAND_nWE_Pin | NAND_nRE_Pin | 0xFF);
    __NOP();
    __NOP();
    return data;
}

void nand_write_byte(uint8_t data)
{
    //
    // CLE=0, ALE=0, nWE=0, nRE=1
    //
    LL_GPIO_WriteOutputPort(NAND_D0_GPIO_Port, NAND_nRE_Pin | data);
    __NOP();
    __NOP();
    __NOP();
    __NOP();
    __NOP();
    __NOP();
    __NOP();
    __NOP();
    __NOP();
    __NOP();
    //
    // CLE=0, ALE=0, nWE=1, nRE=1
    //
    LL_GPIO_WriteOutputPort(NAND_D0_GPIO_Port,
            NAND_nWE_Pin | NAND_nRE_Pin | data);
    __NOP();
    __NOP();
}

static void fx_nand_cmd_reset(NAND_Chip chipselect)
{
    nand_chip_enable(chipselect);

    nand_set_dir_output();

    nand_write_cmd(NAND_CMD_RESET);

    while (!nand_get_ready(chipselect))
    {
        ;
    }
}

//
// Set up the interface and initiate RESET cmd
//
void nand_init(void)
{
    nand_bus_configure();

    LL_EXTI_InitTypeDef EXTI_InitStruct = { 0 };
    LL_SYSCFG_SetEXTISource(LL_SYSCFG_EXTI_PORTC, LL_SYSCFG_EXTI_LINE13);
    EXTI_InitStruct.Line_0_31 = LL_EXTI_LINE_13;
    EXTI_InitStruct.LineCommand = ENABLE;
    EXTI_InitStruct.Mode = LL_EXTI_MODE_IT;
    EXTI_InitStruct.Trigger = LL_EXTI_TRIGGER_RISING;
    LL_EXTI_Init(&EXTI_InitStruct);

    fx_sem_init(&nand_ready_sem, 0, 1, FX_SYNC_POLICY_FIFO);

    nand_set_dir_input();
    LL_GPIO_SetOutputPin(NAND_nWP_GPIO_Port, NAND_nWP_Pin);
    LL_GPIO_WriteOutputPort(NAND_D0_GPIO_Port, NAND_nWE_Pin | NAND_nRE_Pin);

    fx_nand_cmd_reset(Chip0_Select);
}
