/**
  ******************************************************************************
  *  @file   hid.c
  *  @brief  USB HID device example for MDR32 family
  *
  ******************************************************************************
    LPCUSB, an USB device driver for LPC microcontrollers
    Copyright (C) 2006 Bertrik Sikken (bertrik@sikken.nl)
    Copyright (c) 2020 Eremex Ltd.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
       notice, this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.
    3. The name of the author may not be used to endorse or promote products
       derived from this software without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
    IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
    OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
    IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
    NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
    THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *****************************************************************************/

#include <stdint.h>
#include <FXRTOS.h>

#include "usbdevapi.h"
#include "MDR32Fx.h"

enum
{
	INTR_IN_EP = 0x81,
	MAX_PACKET_SIZE = 64,
	REPORT_SIZE = 4
};

static uint8_t report[] = {0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07};
static int idle_rate;
static int frame_num;

//!
//! Mouse report descriptor
//!
static uint8_t report_desc[] =
{
	0x05,    0x01,     //!< Usage Page (Generic Desktop)
	0x09,    0x02,     //!< Usage (Mouse)
	0xA1,    0x01,     //!< Collection (Application)
	0x09,    0x01,     //!< Usage (Pointer)
	0xA1,    0x00,     //!< Collection (Physical)
	0x05,    0x09,     //!< Usage Page (Buttons)
	0x19,    0x01,     //!< Usage Minimum (01)
	0x29,    0x01,     //!< Usage Maximum (01)
	0x15,    0x00,     //!< Logical Minimum (0)
	0x25,    0x01,     //!< Logical Maximum (1)
	0x95,    0x01,     //!< Report Count (1)
	0x75,    0x01,     //!< Report Size (1)
	0x81,    0x02,     //!< Input (Data, Variable, Absolute)
	0x95,    0x01,     //!< Report Count (1)
	0x75,    0x07,     //!< Report Size (7)
	0x81,    0x01,     //!< Input (Constant) for padding
	0x05,    0x01,     //!< Usage Page (Generic Desktop)
	0x09,    0x30,     //!< Usage (X)
	0x09,    0x31,     //!< Usage (Y)
	0x15,    0x81,     //!< Logical Minimum (-127)
	0x25,    0x7F,     //!< Logical Maximum (127)
	0x75,    0x08,     //!< Report Size (8)
	0x95,    0x02,     //!< Report Count (2)
	0x81,    0x06,     //!< Input (Data, Variable, Relative)
	0xC0,//!< End Collection (Physical)
	0xC0 //!< End Collection (Application)
};

static const uint8_t abDescriptors[] =
{
	0x12,                   
	DESC_DEVICE,			//!< Device descriptor
	LE_WORD(0x0110),        //!< bcdUSB
	0x00,                   //!< bDeviceClass
	0x00,                   //!< bDeviceSubClass
	0x00,                   //!< bDeviceProtocol
	MAX_PACKET_SIZE0,       //!< bMaxPacketSize
	LE_WORD(0xFFFF),        //!< idVendor
	LE_WORD(0x0001),        //!< idProduct
	LE_WORD(0x0100),        //!< bcdDevice
	0x01,                   //!< iManufacturer
	0x02,                   //!< iProduct
	0x03,                   //!< iSerialNumber
	0x01,                   //!< bNumConfigurations

	0x09,                   
	DESC_CONFIGURATION,		//!< configuration
	LE_WORD(0x22),          //!< wTotalLength
	0x01,                   //!< bNumInterfaces
	0x01,                   //!< bConfigurationValue
	0x00,                   //!< iConfiguration
	0x80,                   //!< bmAttributes
	0x32,                   //!< bMaxPower

	0x09,                   
	DESC_INTERFACE,			//!< interface
	0x00,                   //!< bInterfaceNumber
	0x00,                   //!< bAlternateSetting
	0x01,                   //!< bNumEndPoints
	0x03,                   //!< bInterfaceClass = HID
	0x00,                   //!< bInterfaceSubClass
	0x00,                   //!< bInterfaceProtocol
	0x00,                   //!< iInterface

	0x09,                   //!< HID descriptor
	DESC_HID_HID,           //!< bDescriptorType = HID
	LE_WORD(0x0110),        //!< bcdHID
	0x00,                   //!< bCountryCode
	0x01,                   //!< bNumDescriptors = report
	DESC_HID_REPORT,        //!< bDescriptorType
	LE_WORD(sizeof(report_desc)),

	0x07,
	DESC_ENDPOINT,			//!< EP descriptor
	INTR_IN_EP,             //!< bEndpointAddress
	0x03,                   //!< bmAttributes = INT
	LE_WORD(MAX_PACKET_SIZE),//!< wMaxPacketSize
	10,                     //!< bInterval

	0x04,   
	DESC_STRING,			//!< string descriptors
	LE_WORD(0x0409),

	0x0E,
	DESC_STRING,			//!< manufacturer string
	'L', 0, 'P', 0, 'C', 0, 'U', 0, 'S', 0, 'B', 0,

	0x12,
	DESC_STRING,			//!< product string
	'P', 0, 'r', 0, 'o', 0, 'd', 0, 'u', 0, 'c', 0, 't', 0, 'X', 0,

	0x12,
	DESC_STRING,			//!< serial number string
	'D', 0, 'E', 0, 'A', 0, 'D', 0, 'C', 0, '0', 0, 'D', 0, 'E', 0,
	
	0						//!< terminator
};

//!
//! HID class request handler
//! @param [in] ctx Control endpoint context
//! @param [out] true in case of success else false
//!
static bool
hid_class_req_handle(usb_ctrl_context_t* ctx)
{
	setup_packet_t* psetup = &ctx->setup;
	switch (psetup->bRequest)
	{
	case HID_GET_IDLE:
		DBG("GET IDLE, val=%X, idx=%X\n", psetup->wValue, psetup->wIndex);
		ctx->pbdata[0] = (idle_rate / 4) & 0xFF;
		ctx->transfer_size = 1;
		break;

	case HID_SET_IDLE:
		DBG("SET IDLE, val=%X, idx=%X\n", psetup->wValue, psetup->wIndex);
		idle_rate = ((psetup->wValue >> 8) & 0xFF) * 4;
		break;

	default:
		DBG("Unhandled class %X\n", psetup->bRequest);
		return false;
	}
	return true;
}

//!
//! Standard request handler for HID devices.
//! This function tries to service any HID specific requests.
//! @param [in] ctx Control endpoint context
//! @param [out] true in case of success else false
//!
static bool
hid_std_req_handle(usb_ctrl_context_t* ctx)
{
	setup_packet_t* const psetup = &ctx->setup;

	if ((psetup->bmRequestType == 0x81) && (psetup->bRequest == REQ_GET_DESCRIPTOR))
	{

		const uint8_t type = GET_DESC_TYPE(psetup->wValue);
		switch (type)
		{

		case DESC_HID_REPORT:
			ctx->pbdata = report_desc;
			ctx->transfer_size = sizeof(report_desc);
			break;

		case DESC_HID_HID:
		case DESC_HID_PHYSICAL:
		default:
			//
			// search descriptor space
			//
			return get_descriptor(ctx);
		}

		return true;
	}
	return false;
}

//!
//! Handle mass storage class request
//! @param [in] USB frame number
//!
static void
HandleFrame(uint16_t frame)
{
	static int frame_count;

	frame_num++;
	if ((frame_num > 1000))
	{
		//
		// send report (dummy data)
		//
		report[0] = (frame_count >> 8) & 0xFF;
		report[1] = (frame_count) & 0xFF;
		frame_count++;
		udc_ep_write(INTR_IN_EP, report, REPORT_SIZE);
		frame_num = 0;
	}
}

fx_sem_t g_usb_sem;
static fx_dpc_t dpc;

//!
//! Deferred function to call from the ISR
//!
static void
usb_dpc_handler(fx_dpc_t* dpc, void* arg)
{
	//
	// Post semaphore and wake the waiting thread up.
	//
	fx_sem_post(&g_usb_sem);
}

//!
//! This function is called by the kernel as a result of any interrupt.
//! HAL or CPU interfaces should be used in order to distinct one vector
//! from another. This demo uses HAL function to obtain the vector.
//!
void
fx_intr_handler(void)
{
	unsigned int vector = hal_intr_get_current_vect();
	//
	// Check whether the vector is ours
	//
	switch (vector)
	{
	case 2:
		NVIC_DisableIRQ(USB_IRQn);
		fx_dpc_request(&dpc, usb_dpc_handler, NULL);
		break;

	default:
		break;
	}
}

int
device_main(void* arg)
{
	(void)arg;

	//
	// Initialise OS sync objects
	//
	fx_sem_init(&g_usb_sem, 0, 1, FX_SYNC_POLICY_FIFO);
	fx_sem_post(&g_usb_sem);
	fx_dpc_init(&dpc);

	DBG("Initialising USB stack\n");

	//
	// Initialise USB device stack
	//
	usb_ctrl_context_t* pctrl = ud_init();
	ud_register_desc(abDescriptors, pctrl);
	ud_register_custom_req_handler(hid_std_req_handle, pctrl);
	ud_register_req_handler(REQTYPE_TYPE_CLASS, hid_class_req_handle, pctrl);
	udc_ep_register_isr(INTR_IN_EP, NULL, NULL);
	udc_dev_register_frame_handler(HandleFrame);

	DBG("Starting USB communication\n");

	//
	// Wait for wake signal from the interrupt handler,
	// service the irq, and wait again.
	//
	while (1)
	{
		fx_sem_timedwait(&g_usb_sem, FX_THREAD_INFINITE_TIMEOUT);
		udc_int_handler();
		NVIC_EnableIRQ(USB_IRQn);
	}

	return 0;
}
