/**
  ******************************************************************************
  *  @file   msc.c
  *  @brief  USB MSC device example for MDR32 family
  *
  ******************************************************************************
    LPCUSB, an USB device driver for LPC microcontrollers
    Copyright (C) 2006 Bertrik Sikken (bertrik@sikken.nl)
    Copyright (c) 2020 Eremex Ltd.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
       notice, this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.
    3. The name of the author may not be used to endorse or promote products
       derived from this software without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
    IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
    OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
    IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
    NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
    THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *****************************************************************************/
#include <stdint.h>
#include <FXRTOS.h>

#include "msc_bot.h"
#include "MDR32Fx.h"

#define MAX_PACKET_SIZE		64

//!
//! Acceptable values depend on hardware and UDC driver
//!
#define MSC_OUT_EP     0x02
#define MSC_IN_EP      0x81

static const uint8_t abDescriptors[] =
{
	0x12,
	DESC_DEVICE,            //!< device descriptor
	LE_WORD(0x0200),        //!< bcdUSB
	0x00,                   //!< bDeviceClass
	0x00,                   //!< bDeviceSubClass
	0x00,                   //!< bDeviceProtocol
	MAX_PACKET_SIZE0,       //!< bMaxPacketSize
	LE_WORD(0xFFFF),        //!< idVendor
	LE_WORD(0x0003),        //!< idProduct
	LE_WORD(0x0100),        //!< bcdDevice
	0x01,                   //!< iManufacturer
	0x02,                   //!< iProduct
	0x03,                   //!< iSerialNumber
	0x01,                   //!< bNumConfigurations

	0x09,
	DESC_CONFIGURATION,     //!< configuration descriptor
	LE_WORD(32),            //!< wTotalLength
	0x01,                   //!< bNumInterfaces
	0x01,                   //!< bConfigurationValue
	0x00,                   //!< iConfiguration
	0xC0,                   //!< bmAttributes
	0x32,                   //!< bMaxPower

	0x09,
	DESC_INTERFACE,         //!< interface
	0x00,                   //!< bInterfaceNumber
	0x00,                   //!< bAlternateSetting
	0x02,                   //!< bNumEndPoints
	0x08,                   //!< bInterfaceClass = mass storage
	0x06,                   //!< bInterfaceSubClass = transparent SCSI
	0x50,                   //!< bInterfaceProtocol = BOT
	0x00,                   //!< iInterface

	0x07,
	DESC_ENDPOINT,          //!< EP
	MSC_IN_EP,         //!< bEndpointAddress
	0x02,                   //!< bmAttributes = bulk
	LE_WORD(MAX_PACKET_SIZE),//!< wMaxPacketSize
	0x00,                   //!< bInterval

	0x07,
	DESC_ENDPOINT,          //!< EP
	MSC_OUT_EP,        //!< bEndpointAddress
	0x02,                   //!< bmAttributes = bulk
	LE_WORD(MAX_PACKET_SIZE),//!< wMaxPacketSize
	0x00,                   //!< bInterval

	0x04,
	DESC_STRING,            //!< string descriptors
	LE_WORD(0x0409),        //!< wLangid (U.S. English)

	0x0E,
	DESC_STRING,
	'E', 0, 'R', 0, 'E', 0, 'M', 0, 'E', 0, 'X', 0, //!< Manufacturer

	0x12,
	DESC_STRING,
	'P', 0, 'r', 0, 'o', 0, 'd', 0, 'u', 0, 'c', 0, 't', 0, 'X', 0, //!< Product

	0x1A,
	DESC_STRING,
	'D', 0, 'E', 0, 'A', 0, 'D', 0, 'C', 0, '0', 0, 'D', 0, 'E', 0,
	'C', 0, 'A', 0, 'F', 0, 'E', 0, //!< Serial No

	0 //!< terminating zero
};

msc_context_t msc_ctx =
{
	.in_ep = MSC_IN_EP,
	.out_ep = MSC_OUT_EP,
};

fx_sem_t g_usb_sem;
static fx_dpc_t dpc;

//!
//! Handle mass storage class request
//! @param [in] ctx Control endpoint context
//! @param [out] true in case of success else false
//!
static bool
msc_req_handle(usb_ctrl_context_t* ctx)
{
	setup_packet_t* psetup = &ctx->setup;

	if (psetup->wIndex != 0)
	{
		DBG("Invalid idx %X\n", psetup->wIndex);
		return false;
	}
	if (psetup->wValue != 0)
	{
		DBG("Invalid val %X\n", psetup->wValue);
		return false;
	}

	switch (psetup->bRequest)
	{
	//
	// get max LUN (one LUN)
	//
	case 0xFE:
		ctx->pbdata[0] = 0;
		ctx->transfer_size = 1;
		break;
	//
	// MSC reset
	//
	case 0xFF:
		if (psetup->wLength > 0)
		{
			return false;
		}
		msc_reset(&msc_ctx);
		break;

	default:
		DBG("Unhandled class\n");
		return false;
	}
	return true;
}

//!
//! Deferred function to call from the ISR
//!
static void
usb_dpc_handler(fx_dpc_t* dpc, void* arg)
{
	//
	// Post semaphore and wake the waiting thread up.
	//
	fx_sem_post(&g_usb_sem);
}

//!
//! This function is called by the kernel as a result of any interrupt.
//! HAL or CPU interfaces should be used in order to distinct one vector
//! from another. This demo uses HAL function to obtain the vector.
//!
void
fx_intr_handler(void)
{
	unsigned int vector = hal_intr_get_current_vect();
	//
	// Check whether the vector is ours
	//
	switch (vector)
	{
	case USB_IRQn:
		NVIC_DisableIRQ(USB_IRQn);
		fx_dpc_request(&dpc, usb_dpc_handler, NULL);
		break;

	default:
		break;
	}
}

//!
//! MSC device thread function
//! @arg arg unused
//!
void
device_main(void* arg)
{
	(void)arg;

	//
	// Initialise storage block device
	//
	blockdev_init();

	//
	// Initialise OS sync objects
	//
	fx_sem_init(&g_usb_sem, 0, 1, FX_SYNC_POLICY_FIFO);
	fx_sem_post(&g_usb_sem);
	fx_dpc_init(&dpc);

	DBG("Initialising USB stack\n");

	//
	// Initialise USB device stack
	//
	usb_ctrl_context_t* pctrl = ud_init();
	ud_register_desc(abDescriptors, pctrl);
	ud_register_req_handler(REQTYPE_TYPE_CLASS, msc_req_handle, pctrl);
	udc_ep_register_isr(MSC_IN_EP, (ep_int_handler_t*)msc_bulk_in, &msc_ctx);
	udc_ep_register_isr(MSC_OUT_EP, (ep_int_handler_t*)msc_bulk_out, &msc_ctx);

	DBG("Starting USB communication\n");

	//
	// Wait for wake signal from the interrupt handler,
	// service the irq, and wait again.
	//
	while (1)
	{
		fx_sem_timedwait(&g_usb_sem, FX_THREAD_INFINITE_TIMEOUT);
		udc_int_handler();
		NVIC_EnableIRQ(USB_IRQn);
	}
}
