/**
  ******************************************************************************
  *  @file   udc.c
  *  @brief  USB device hardware layer for Milandr 1986ВЕ9x, 1986BEx MCUs
  *
  ******************************************************************************
  *  Copyright (C) 2020 Eremex
  *  All rights reserved.
  *****************************************************************************/

#include <assert.h>
#include <stdint.h>
#include <stdbool.h>

#include <MDR32F9Qx_usb.h>
#include <MDR32F9Qx_rst_clk.h>

#include "usbdevapi.h"

//!
//! Installed device interrupt handler
//!
static dev_int_handler_t* pfn_dev_int_handler = NULL;

//!
//! Installed frame interrupt handlers
//!
static dev_frame_handler_t* pfn_frame_handler = NULL;

//!
//! Installed endpoint interrupt handlers
//!
static ep_int_handler_t* pfn_ep_int_handlers[Num_USB_EndPoints];
static void* ep_handler_data[Num_USB_EndPoints];

static uint8_t new_address = 0;
static int last_req_error;
static int in_sent;

//!
//! Convert from endpoint address to endpoint index
//! @param [in] ep  Endpoint number (address)
//! @param [out]    Endpoint index in handler list
//!
static inline uint8_t
EP2IDX(uint8_t ep)
{
	uint8_t idx = ep & 0x3;
	ASSERT(idx <= 0x3);
	return idx;
}

//!
//! Convert from endpoint index to endpoint address
//! In this driver it doesn't work for idx 0 which have dual direcion endpoint,
//! need additional check
//! @param [in] ep  Endpoint index
//! @param [out]    Endpoint address
//!
uint8_t static inline
IDX2EP(uint8_t idx)
{
	ASSERT(idx <= 0x3);
	return idx | ((idx & 0x1) << 7);
}

//!
//! Enables or disables an endpoint.
//! @param [in] idx     Endpoint index
//! @param [in] enable  true to enable, false to disable
//!
static void
udc_ep_enable(int idx, bool enable)
{
	uint32_t reg = USB_GetSEPxCTRL(idx);
	USB_SetSEPxCTRL(idx, reg | USB_SEPx_CTRL_EPEN_Enable);
}

//!
//! Set endpoint as ready to receive or transmit packet
//! @param [in] idx     Endpoint index
//! @param [in] enable  true to enable, false to disable
//!
static void
udc_ep_set_ready(int idx, bool enable)
{
	uint32_t reg = USB_GetSEPxCTRL(idx);
	USB_SetSEPxCTRL(idx, reg | (enable ? USB_SEPx_CTRL_EPRDY_Ready : 0));
}

static int
udc_ep_get_ready(int idx)
{
	return USB_SEPx_CTRL_EPRDY_Ready & USB_GetSEPxCTRL(idx);
}

//!
//! Configures an endpoint and enables it
//! @param [in] ep              Endpoint number
//! @param [in] max_packet_sz   Maximum packet size for this EP
//!
void
udc_ep_configure(uint8_t ep, uint16_t max_packet_sz)
{
	const int idx = EP2IDX(ep);

	USB_SetSEPxCTRL(idx, USB_SEPx_CTRL_EPDATASEQ_Data0);
	udc_ep_enable(idx, true);

	//
	// Reset RX counter and TX queue
	//
	USB_SetSEPxRXFC(idx, 1);
	USB_SetSEPxTXFDC(idx, 1);

	if (REQTYPE_GET_DIR(ep) == REQTYPE_DIR_TO_DEVICE)
	{
		udc_ep_set_ready(idx, true);
	}
}

//!
//! Registers an endpoint interrupt event callback
//! @param [in] ep              Endpoint number
//! @param [in] fn_handler      Callback function
//! @param [in] ctx      Endpoint private context for use in callback
//!
void
udc_ep_register_isr(uint8_t ep, ep_int_handler_t* fn_handler, void* ctx)
{
	const int idx = EP2IDX(ep);
	ASSERT(idx < Num_USB_EndPoints);

	udc_ep_configure(ep, 64);
	//
	// Add handler and its context to lists
	//
	pfn_ep_int_handlers[idx] = fn_handler;
	ep_handler_data[idx] = ctx;

	DBG("Registered handler for EP 0x%x\n", ep);
}

//!
//! Registers an device status callback
//! @param [in] fn_handler  Callback function
//!
void
udc_dev_register_isr(dev_int_handler_t* fn_handler)
{
	pfn_dev_int_handler = fn_handler;

	DBG("Registered handler for device status\n");
}

//!
//! Registers specific frame callback if needed
//! @param [in] fn_handler  Callback function
//!
void
udc_dev_register_frame_handler(dev_frame_handler_t* handler)
{
	pfn_frame_handler = handler;

	DBG("Registered handler for frame\n");
}

//!
//! Sets the USB address.
//! @param [in] addr        Device address to set
//!
void
udc_dev_set_address(uint8_t addr)
{
	//
	// The address is applied in udc_int_handler() when new frame comes;
	//
	new_address = addr;
}

//!
//! Gets the status from a specific endpoint.
//! @param [in] ep      Endpoint number
//! @return Endpoint status byte (containing EP_STATUS_xxx bits)
//!
uint8_t
udc_ep_get_status(uint8_t ep)
{
	const int idx = EP2IDX(ep);

	const bool is_setup = USB_GetSEPxTS(idx) == USB_SEPx_TS_SCTTYPE_Setup;
	const bool nak = (USB_GetSEPxSTS(idx) & USB_SEPx_STS_SCNAKSENT_Set) != 0;
	const bool crc_err = (USB_GetSEPxSTS(idx) & USB_SEPx_STS_SCCRCERR_Set) != 0;
	const bool bs_err = (USB_GetSEPxSTS(idx) & USB_SEPx_STS_SCBSERR_Set) != 0;
	const bool stall = (USB_GetSEPxCTRL(idx) & USB_SEPx_CTRL_EPSSTALL_Reply) != 0;

	uint8_t stat = (is_setup) ? EP_STATUS_SETUP : EP_STATUS_DATA;
	if (nak)
	{
		stat |= EP_STATUS_NACKED;
	}
	if (crc_err || bs_err)
	{
		stat |= EP_STATUS_ERROR;
	}
	if (stall)
	{
		stat |= EP_STATUS_STALLED;
	}

	return stat;
}

//!
//! Sets the stalled property of an endpoint
//! @param [in] ep      Endpoint number
//! @param [in] stall   true to stall, false to unstall
//!
void
udc_ep_stall(uint8_t ep, bool stall)
{
	const int idx = EP2IDX(ep);
	if (stall)
	{
		MDR_USB->USB_SEP[idx].CTRL |=
		    USB_SEPx_CTRL_EPSSTALL_Reply |  USB_SEPx_CTRL_EPRDY_Ready;
		last_req_error = 1;
	}
	else
	{
		MDR_USB->USB_SEP[idx].CTRL &= ~USB_SEPx_CTRL_EPSSTALL_Reply;
		last_req_error = 0;
	}
}

//!
//! Writes data to an endpoint buffer
//! @param [in] ep      Endpoint number
//! @param [in] pbuf    Endpoint data
//! @param [in] len Number of bytes to write
//! @return true if the data was successfully written or <0 in case of error.
//!
int
udc_ep_write(uint8_t ep, uint8_t* pbuf, int len)
{
	const int idx = EP2IDX(ep);

	USB_SetSEPxTXFDC(idx, 1);
	for (unsigned i = 0; i < len; i++)
	{
		USB_SetSEPxTXFD(idx, pbuf[i]);
	}

	udc_ep_set_ready(idx, true);
	in_sent = 1;

	return len;
}

//!
//! Reads data from an endpoint buffer
//! @param [in] ep      Endpoint number
//! @param [in] pbuf    Endpoint data
//! @param [in] max_len Maximum number of bytes to read
//! @return the number of bytes available in the EP (possibly more than max_len)
//! or <0 in case of error.
//!
int
udc_ep_read(uint8_t ep, uint8_t* pbuf, int max_len)
{
	const int idx = EP2IDX(ep);
	uint32_t len = USB_GetSEPxRXFDC(idx);
	len = (len < max_len) ? len : max_len;

	for (size_t i = 0; i < len; i++)
	{
		pbuf[i] = (uint8_t)USB_GetSEPxRXFD(idx);
	}
	USB_SetSEPxRXFC(idx, 1);

	return len;
}

//!
//! Sets the 'configured' state.
//! Dummy function for this type of controller
//! @param [in] configured  If true, configure device, else unconfigure
//!
void
udc_dev_configure(bool configured)
{
	;
}

//!
//! Sets the device 'reset' state.
//!
void
udc_dev_reset()
{
	USB_SetSA(0);
}

//!
//! Sets the DATAx sequence for the endpoint.
//! @param [in] ep Endpoint number
//! @param [in] seq Accept values "1" for DATA1, "0" for DATA0
//!
void
udc_ep_set_dataseq(uint8_t ep, bool seq)
{
	const int idx = EP2IDX(ep);
	uint32_t tmpreg = MDR_USB->USB_SEP[idx].CTRL;
	tmpreg = seq ?
	         (tmpreg | USB_SEPx_CTRL_EPDATASEQ_Data1) :
	         (tmpreg & ~USB_SEPx_CTRL_EPDATASEQ_Data1);
	MDR_USB->USB_SEP[idx].CTRL = tmpreg;
}

//!
//! USB interrupt handler
//!
void
udc_int_handler(void)
{
	//
	// Read device interrupt status
	//
	const uint32_t events = USB_GetSIS();

	//
	// SOF received
	//
	if (events & USB_SIS_SCSOFREC)
	{
		MDR_USB->SIS |= USB_SIS_SCSOFREC;
		if (new_address)
		{
			USB_SetSA(new_address);
			new_address = 0;
		}
		if (last_req_error)
		{
			udc_ep_stall(0, false);
		}
		if (pfn_frame_handler != NULL)
		{
			pfn_frame_handler(0);
		}
	}
	//
	// Reset
	//
	if (events & USB_SIS_SCRESETEV)
	{
		MDR_USB->SIS |= USB_SIS_SCRESETEV;
		pfn_dev_int_handler(DEV_STATUS_RESET);
	}
	//
	// Resume
	//
	if (events & USB_SIS_SCRESUME)
	{
		MDR_USB->SIS |= USB_SIS_SCRESUME;
		pfn_dev_int_handler(DEV_STATUS_RESUME);
	}
	//
	// Transfer done
	//
	if (events & USB_SIS_SCTDONE)
	{
		MDR_USB->SIS |= USB_SIS_SCTDONE;
		for (size_t i = USB_EP0; i < Num_USB_EndPoints; i++)
		{
			const bool disabled = (USB_GetSEPxCTRL(i) & USB_SEP_CTRL_EPEN) == 0;
			const bool ready_for_transfer = udc_ep_get_ready(i);
			if (disabled || ready_for_transfer)
			{
				continue;
			}

			uint8_t last_xmit_type = USB_GetSEPxTS(i);
			uint8_t ep = IDX2EP(i);
			//
			// Special case to detect control IN endpoint
			// because ep 0 is bidirectional)
			//
			if (ep == 0 && (last_xmit_type & USB_SEPx_TS_SCTTYPE_In))
			{
				ep = 0x80;
			}
			//
			// In any case status on SETUP must be DATA1
			//
			if (last_xmit_type == USB_SEPx_TS_SCTTYPE_Setup)
			{
				udc_ep_set_dataseq(0x0, 1);
			}
			//
			// Toggle DATAx if OUT DATAx packet was received from host
			//
			else if (last_xmit_type == USB_SEPx_TS_SCTTYPE_Outdata)
			{
				USB_SEPxToggleEPDATASEQ(i);
			}
			else if ((REQTYPE_GET_DIR(ep) == REQTYPE_DIR_TO_HOST) && in_sent)
			{
				if (USB_SEP_STS_SCACKRXED & USB_GetSEPxSTS(i))
				{
					//
					// Toggle DATAx after packet was sent
					//
					USB_SEPxToggleEPDATASEQ(i);
					in_sent = 0;
				}
			}

			//
			// Call EP handler if nonzero data transfered in any direction
			//
			size_t rx_len = USB_GetSEPxRXFDC(i);
			if (rx_len || (REQTYPE_GET_DIR(ep) == REQTYPE_DIR_TO_HOST))
			{
				const uint8_t stat = udc_ep_get_status(i);
				if (pfn_ep_int_handlers[i] != NULL)
				{
					pfn_ep_int_handlers[i](ep, stat, ep_handler_data[i]);
				}
			}

			if (REQTYPE_GET_DIR(ep) != REQTYPE_DIR_TO_HOST)
			{
				udc_ep_set_ready(i, true);
			}
		}
	}
}

//!
//! Initialises the USB hardware
//!  @return true if the hardware was successfully initialised
//!
bool
udc_init(void)
{
	//
	// Enables the CPU_CLK clock on USB
	//
	RST_CLK_PCLKcmd(RST_CLK_PCLK_USB, ENABLE);
	//
	// Device layer initialization
	//
	USB_Clock_TypeDef USB_Clock_InitStruct;
	USB_Clock_InitStruct.USB_USBC1_Source = USB_C1HSEdiv2;
	USB_Clock_InitStruct.USB_PLLUSBMUL    = USB_PLLUSBMUL12;
	//
	// Set USB clocks perform controller reset
	//
	USB_BRGInit(&USB_Clock_InitStruct);
	USB_Reset();
	//
	// Set pulls and Device mode
	//
	USB_SetHSCR(USB_HSCR_HOST_MODE_Device);
	USB_SetHSCR(USB_HSCR_DP_PULLUP_Set);
	//
	// Set speed, polarity and enable end points
	//
	USB_SetSC(USB_SC_SCFSR_12Mb | USB_SC_SCFSP_Full | USB_SC_SCGEN_Set);
	//
	// Enable all USB interrupts
	//
	USB_SetSIM(USB_SIS_Msk);
	//
	// USB Power On
	//
	USB_SetHSCR(USB_HSCR_EN_RX_Set | USB_HSCR_EN_TX_Set);
	//
	// Setup Address 0
	//
	USB_SetSA(0);

	return true;
}
