/**
  ******************************************************************************
  *  @file   usbdevapi.h
  *  @brief  USB device stack API
  *  Contains:
  *     - USB device low level API (HAL)
  *     - USB device definitions for user application
  *     - USB device stack interface
  *     - Block device interface for Mass storage class
  *
  ******************************************************************************
    LPCUSB, an USB device driver for LPC microcontrollers
    Copyright (C) 2006 Bertrik Sikken (bertrik@sikken.nl)
    Copyright (c) 2020 Eremex Ltd.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
       notice, this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.
    3. The name of the author may not be used to endorse or promote products
       derived from this software without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
    IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
    OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
    IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
    NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
    THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *****************************************************************************/

#ifndef _USBDEVAPI_H_
#define _USBDEVAPI_H_

#include <stdint.h>
#include <stdbool.h>
#include <stddef.h>

//!
//! Service macros
//!
#define MIN(x,y)    ((x)<(y)?(x):(y))
#define MAX(x,y)    ((x)>(y)?(x):(y))
#define LE_WORD(x)	((x)&0xFF),((x)>>8)

//!
//! USB device low level API (HAL)
//!

//!
//! USB configuration
//!
enum
{
	MAX_PACKET_SIZE0 = 64   //!< maximum packet size for EP 0
};

//!
//! USB device debugging functions
//!
#ifndef DEBUG_USB
#define DBG(...)
#define ASSERT(x)
#endif

//!
//! USB hardware interface
//!

//!
//! endpoint status sent through callback
//!
enum
{
	EP_STATUS_DATA    = (1 << 0), //!< EP has data
	EP_STATUS_STALLED = (1 << 1), //!< EP is stalled
	EP_STATUS_SETUP   = (1 << 2), //!< EP received setup packet
	EP_STATUS_ERROR   = (1 << 3), //!< EP data was overwritten by setup packet
	EP_STATUS_NACKED  = (1 << 4), //!< EP sent NAK
};

//!
//! device status sent through callback
//!
enum
{
	DEV_STATUS_CONNECT  = (1 << 0), //!< device just got connected
	DEV_STATUS_RESUME   = (1 << 1), //!< tells device to wake up
	DEV_STATUS_SUSPEND  = (1 << 2), //!< device entered suspend state
	DEV_STATUS_RESET    = (1 << 4) //!< device just got reset
};

bool udc_init(void);
void udc_int_handler(void);

void udc_dev_set_address(uint8_t addr);
void udc_dev_configure(bool configured);

//!
//! endpoint operations
//!
void udc_ep_configure(uint8_t ep, uint16_t max_pkt_sz);
int  udc_ep_read(uint8_t ep, uint8_t* pbuf, int max_len);
int  udc_ep_write(uint8_t ep, uint8_t* pbuf, int len);
void udc_ep_stall(uint8_t ep, bool stall);
uint8_t   udc_ep_get_status(uint8_t ep);

//!
//! Endpoint interrupt handler callback
//!
typedef void (ep_int_handler_t)(uint8_t ep, uint8_t ep_status, void* ep_ctx);
void udc_ep_register_isr(uint8_t ep, ep_int_handler_t* pfn_handler, void* ctx);

//!
//! Device status handler callback
//!
typedef void (dev_int_handler_t)(uint8_t dev_status);
void udc_dev_register_isr(dev_int_handler_t* pfn_handler);

//!
//! Device SOF handler callback
//!
typedef void (dev_frame_handler_t)(uint16_t frame);
void udc_dev_register_frame_handler(dev_frame_handler_t* handler);

void udc_dev_reset(void);
void udc_ep_set_dataseq(uint8_t ep, bool seq);

//!
//! USB device high level API
//!
enum
{
	SETUP_PKT_SZ = 8,   	//!< Every Setup packet has eight bytes
	MAX_CONTROL_SIZE = 128, //!< maximum total size of control transfer data
	MAX_REQ_TYPES = 4		//!< standard, class, vendor, reserved
};

//!
//! setup packet definitions
//!
typedef struct
{
	uint8_t bmRequestType;          //<! characteristics of the specific request
	uint8_t bRequest;               //<! specific request
	uint16_t    wValue;             //<! request specific parameter
	uint16_t    wIndex;             //<! request specific parameter
	uint16_t    wLength;            //<! length of data transfered in data phase
} setup_packet_t;

//!
//! Setup packet bmRequestType direction bitfield
//!
enum
{
	REQTYPE_DIR_TO_DEVICE = 0,
	REQTYPE_DIR_TO_HOST = 1
};

#define REQTYPE_GET_DIR(x)      (((x)>>7)&0x01)
#define REQTYPE_GET_TYPE(x)     (((x)>>5)&0x03)
#define REQTYPE_GET_RECIP(x)    ((x)&0x1F)

//!
//! Setup packet bmRequestType type bitfield
//!
enum
{
	REQTYPE_TYPE_STANDARD = 0,
	REQTYPE_TYPE_CLASS = 1,
	REQTYPE_TYPE_VENDOR = 2,
	REQTYPE_TYPE_RESERVED = 3
};

//!
//! Descriptor types for user descriptor construction
//!
enum
{
	DESC_DEVICE           = 1,
	DESC_CONFIGURATION    = 2,
	DESC_STRING           = 3,
	DESC_INTERFACE        = 4,
	DESC_ENDPOINT         = 5,
	DESC_DEVICE_QUALIFIER = 6,
	DESC_OTHER_SPEED      = 7,
	DESC_INTERFACE_POWER  = 8
};

//!
//! Desctriptor subtypes for HID device
//!
enum
{
	DESC_HID_HID      = 0x21,
	DESC_HID_REPORT   = 0x22,
	DESC_HID_PHYSICAL = 0x23
};

//!
//! Setup packet bmRequestType recipient bitfield
//!
enum
{
	REQTYPE_RECIP_DEVICE = 0,
	REQTYPE_RECIP_INTERFACE = 1,
	REQTYPE_RECIP_ENDPOINT = 2,
	REQTYPE_RECIP_OTHER = 3
};

//!
//! standard requests
//!
enum
{
	REQ_GET_STATUS        = 0x00,
	REQ_CLEAR_FEATURE     = 0x01,
	REQ_SET_FEATURE       = 0x03,
	REQ_SET_ADDRESS       = 0x05,
	REQ_GET_DESCRIPTOR    = 0x06,
	REQ_SET_DESCRIPTOR    = 0x07,
	REQ_GET_CONFIGURATION = 0x08,
	REQ_SET_CONFIGURATION = 0x09,
	REQ_GET_INTERFACE     = 0x0A,
	REQ_SET_INTERFACE     = 0x0B,
	REQ_SYNCH_FRAME       = 0x0C
};

//!
//! class requests HID
//!
enum
{
	HID_GET_REPORT   = 0x01,
	HID_GET_IDLE     = 0x02,
	HID_GET_PROTOCOL = 0x03,
	HID_SET_REPORT   = 0x09,
	HID_SET_IDLE     = 0x0A,
	HID_SET_PROTOCOL = 0x0B
};

//!
//! feature selectors
//!
enum
{
	FEA_ENDPOINT_HALT = 0x00,
	FEA_REMOTE_WAKEUP = 0x01,
	FEA_TEST_MODE     = 0x02
};

//!
//! USB descriptor header
//!
typedef struct
{
	uint8_t bLength;            //!< descriptor length
	uint8_t bDescriptorType;    //!< descriptor type
} device_desc_hdr_t;

#define GET_DESC_TYPE(x)        (((x)>>8)&0xFF)
#define GET_DESC_INDEX(x)       ((x)&0xFF)

//!
//! Device control transfer context
//!
struct _control_context_s;
typedef bool (ud_req_handler_t)(struct _control_context_s* ctrl_context);
typedef struct _control_context_s
{
	//
	// Setup packet
	//
	setup_packet_t setup; //!< packet contents
	//
	// Data stage for control requests
	//
	uint8_t* pbdata; //!< pointer to DATA buffer for request/response
	size_t residue; //!< remaining bytes in DATA buffer
	size_t transfer_size; //!< total length of DATA transfer
	uint8_t data_store[MAX_REQ_TYPES][SETUP_PKT_SZ]; //!< bufs for ctrl requests
	//
	// Device configuration
	//
	uint8_t configuration;      //!< Active configuration
	const uint8_t* descriptors; //!< Descriptor array
	//
	// Control request handlers
	//
	ud_req_handler_t* req_handlers[MAX_REQ_TYPES];//!< request handler callbacks
	ud_req_handler_t* custom_req_handler;//!< custom upper level request handler
} usb_ctrl_context_t;

//!
//! Initialise complete stack, including HW
//!
usb_ctrl_context_t* ud_init(void);

//!
//! Add descriptors to control endpoint's context
//!
void ud_register_desc(const uint8_t* pabDescriptors, usb_ctrl_context_t* ctx);

//!
//! Descriptor handler callback
//!
bool get_descriptor(usb_ctrl_context_t* ctx);

//!
//! Register request handler callback (standard, vendor, class)
//!
void ud_register_req_handler(int type, ud_req_handler_t* pfn_handler,
                             usb_ctrl_context_t* ctx);

//!
//! Register upper level request handler
//!
void ud_register_custom_req_handler(ud_req_handler_t* pfn_handler,
                                    usb_ctrl_context_t* ctx);

#endif
