/*
 *  Copyright (C) JSC EREMEX, 2021.
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are met:
 *  1. Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the copyright holder nor the names of its
 *     contributors may be used to endorse or promote products derived from
 *     this software without specific prior written permission.
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 *  FX-RTOS demo application works on Elbrus 2000 CPU.
 */

#include "FXRTOS.h"

#include <stdint.h>
#include <stdio.h>

#include "fx_printf.h"

extern void hw_timer_setup(void);
extern void hw_pic_init(void);

extern unsigned int hal_intr_get_current_vect(void);

static fx_sem_t sem;

#define AM85C30_RR0 0x00
#define AM85C30_D2 (0x01 << 2)
#define UART_BASE 0x81200040

int
am85c30_putc(int ch)
{
	uint8_t ret;
	do
	{
        e2k_write8(AM85C30_RR0, UART_BASE);
        ret = e2k_read8(UART_BASE);
    }
    while ((ret & AM85C30_D2) == 0);

    e2k_write8(ch, UART_BASE + 0x01);
	return 0;
}

void
fx_intr_handler(void)
{
    unsigned irq = hal_intr_get_current_vect();
}

static void
thread_0(void* arg)
{
    fx_printf("T0 start with arg=%llx\n", arg);
	for (;;)
	{
		fx_printf("0");
		fx_thread_sleep(100);
	}
}

static void
thread_1(void* arg)
{
    fx_printf("T1 start with arg=%llx\n", arg);
	for (;;)
	{
		fx_printf("1");
		fx_sem_post(&sem);
		fx_thread_sleep(200);
	}
}

static void
thread_2(void* arg)
{
    unsigned long i = 0;
    fx_printf("T3 start with arg=%llx\n", arg);
	for (;;)
	{
        fx_sem_wait(&sem, NULL);
		fx_printf("2:%lu\n", i++);
	}
}

uint8_t __attribute__((aligned(4096))) stack0[0x10000];
uint8_t __attribute__((aligned(4096))) stack1[0x10000];
uint8_t __attribute__((aligned(4096))) stack2[0x10000];

void
fx_app_init(void)
{
	static fx_thread_t t0;
	static fx_thread_t t1;
    static fx_thread_t t2;

	fx_printf("APP_INIT\n");

    hw_pic_init();

    hw_timer_setup();

    hw_cpu_intr_enable();

    fx_sem_init(&sem, 0, 1, FX_SYNC_POLICY_DEFAULT);
    hal_intr_set_stack_params(stack0, sizeof(stack0), HAL_INTR_CHAIN_STK_SZ, 0x2000);
    hal_intr_set_stack_params(stack0, sizeof(stack0), HAL_INTR_PROCEDURE_STK_SZ, 0x4000);
    hal_intr_set_stack_params(stack0, sizeof(stack0), HAL_INTR_USER_STK_SZ, 0x4000);
	fx_thread_init(&t0, thread_0, (void*)0xCAFE, 10, stack0, sizeof(stack0), false);
    hal_intr_set_stack_params(stack1, sizeof(stack1), HAL_INTR_CHAIN_STK_SZ, 0x1000);
    hal_intr_set_stack_params(stack1, sizeof(stack1), HAL_INTR_PROCEDURE_STK_SZ, 0x4000);
    hal_intr_set_stack_params(stack1, sizeof(stack1), HAL_INTR_USER_STK_SZ, 0xB000);
	fx_thread_init(&t1, thread_1, (void*)0xBABE, 9, stack1, sizeof(stack1), false);
	hal_intr_set_stack_params(stack2, sizeof(stack2), HAL_INTR_CHAIN_STK_SZ, 0x1000);
    hal_intr_set_stack_params(stack2, sizeof(stack2), HAL_INTR_PROCEDURE_STK_SZ, 0x4000);
    hal_intr_set_stack_params(stack2, sizeof(stack2), HAL_INTR_USER_STK_SZ, 0xB000);
	fx_thread_init(&t2, thread_2, (void*)0xA5A5A5A5A5, 10, stack2, sizeof(stack2), false);
}

int
main(void)
{
	fx_printf("MAIN\n");
	fx_kernel_entry();
}
